from __future__ import annotations

import logging
from collections.abc import Sequence
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Iterable,
    List,
    Optional,
    Self,
    cast,
    overload,
)
from uuid import UUID

import graphene
import graphene_federation
import graphql
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphql import Undefined, UndefinedType
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask.bgtask import ProgressReporter
from ai.backend.common.docker import ImageRef, KernelFeatures, LabelName
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import (
    AgentId,
    DispatchResult,
    ImageAlias,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.models.minilang import EnumFieldItem
from ai.backend.manager.models.minilang.ordering import ColumnMapType, QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import (
    FieldSpecType,
    QueryFilterParser,
)
from ai.backend.manager.models.rbac.context import ClientContext
from ai.backend.manager.models.rbac.permission_defs import ImagePermission
from ai.backend.manager.services.container_registry.actions.clear_images import ClearImagesAction
from ai.backend.manager.services.container_registry.actions.load_all_container_registries import (
    LoadAllContainerRegistriesAction,
)
from ai.backend.manager.services.container_registry.actions.load_container_registries import (
    LoadContainerRegistriesAction,
)
from ai.backend.manager.services.container_registry.actions.rescan_images import RescanImagesAction
from ai.backend.manager.services.image.actions.alias_image import AliasImageAction
from ai.backend.manager.services.image.actions.clear_image_custom_resource_limit import (
    ClearImageCustomResourceLimitAction,
)
from ai.backend.manager.services.image.actions.dealias_image import DealiasImageAction
from ai.backend.manager.services.image.actions.forget_image import (
    ForgetImageAction,
)
from ai.backend.manager.services.image.actions.forget_image_by_id import ForgetImageByIdAction
from ai.backend.manager.services.image.actions.modify_image import (
    ImageModifier,
    ModifyImageAction,
)
from ai.backend.manager.services.image.actions.purge_image_by_id import PurgeImageByIdAction
from ai.backend.manager.services.image.actions.purge_images import (
    PurgedImagesData,
    PurgeImageAction,
    PurgeImageActionResult,
    PurgeImagesActionResult,
)
from ai.backend.manager.services.image.actions.untag_image_from_registry import (
    UntagImageFromRegistryAction,
)
from ai.backend.manager.services.image.types import ImageRefData
from ai.backend.manager.types import OptionalState, TriState

from ...data.image.types import ImageStatus, ImageType
from ...defs import DEFAULT_IMAGE_ARCH
from ...errors.image import ImageNotFound
from ..base import (
    FilterExprArg,
    OrderExprArg,
    batch_multiresult_in_scalar_stream,
    generate_sql_info_for_gql_connection,
)
from ..gql_relay import AsyncNode, Connection, ConnectionResolverResult, ResolvedGlobalID
from ..image import (
    ImageIdentifier,
    ImageLoadFilter,
    ImageRow,
    get_permission_ctx,
)
from ..rbac import ScopeType
from ..user import UserRole
from .base import (
    BigInt,
    ImageRefType,
    KVPair,
    KVPairInput,
    ResourceLimit,
    ResourceLimitInput,
    extract_object_uuid,
)

if TYPE_CHECKING:
    from ..gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "Image",
    "ImageNode",
    "PreloadImage",
    "RescanImages",
    "ForgetImage",
    "ForgetImageById",
    "PurgeImageById",
    "UntagImageFromRegistry",
    "ModifyImage",
    "AliasImage",
    "DealiasImage",
    "ClearImages",
)

_queryfilter_fieldspec: FieldSpecType = {
    "id": ("id", None),
    "name": ("name", None),
    "namespace": ("image", None),
    "tag": ("tag", None),
    "status": (EnumFieldItem("status", ImageStatus), None),
    "project": ("project", None),
    "image": ("image", None),
    "base_image_name": ("image", None),
    "created_at": ("created_at", dtparse),
    "registry": ("registry", None),
    "registry_id": ("registry_id", None),
    "architecture": ("architecture", None),
    "is_local": ("is_local", None),
    "type": (EnumFieldItem("type", ImageType), None),
    "accelerators": ("accelerators", None),
}

_queryorder_colmap: ColumnMapType = {
    "id": ("id", None),
    "name": ("name", None),
    "namespace": ("image", None),
    "tag": ("tag", None),
    "status": ("status", None),
    "project": ("project", None),
    "image": ("image", None),
    "base_image_name": ("image", None),
    "created_at": ("created_at", None),
    "registry": ("registry", None),
    "registry_id": ("registry_id", None),
    "architecture": ("architecture", None),
    "is_local": ("is_local", None),
    "type": ("session_type", None),
    "accelerators": ("accelerators", None),
}

ImageStatusType = graphene.Enum.from_enum(ImageStatus, description="Added in 25.4.0.")
ImageTypeEnum = graphene.Enum.from_enum(ImageType, description="Added in 25.12.0.")


class Image(graphene.ObjectType):
    id = graphene.UUID()
    name = graphene.String(deprecation_reason="Deprecated since 24.12.0. use `namespace` instead")
    namespace = graphene.String(description="Added in 24.12.0.")
    base_image_name = graphene.String(description="Added in 24.12.0.")
    project = graphene.String(description="Added in 24.03.10.")
    humanized_name = graphene.String()
    tag = graphene.String()
    tags = graphene.List(KVPair, description="Added in 24.12.0.")
    version = graphene.String(description="Added in 24.12.0.")
    registry = graphene.String()
    architecture = graphene.String()
    is_local = graphene.Boolean()
    digest = graphene.String()
    labels = graphene.List(KVPair)
    aliases = graphene.List(graphene.String)
    size_bytes = BigInt()
    status = graphene.String(description="Added in 25.4.0.")
    resource_limits = graphene.List(ResourceLimit)
    supported_accelerators = graphene.List(graphene.String)
    installed = graphene.Boolean()
    installed_agents = graphene.List(graphene.String)
    # legacy field
    hash = graphene.String()

    # internal attributes
    raw_labels: dict[str, Any]

    @classmethod
    def populate_row(
        cls,
        ctx: GraphQueryContext,
        row: ImageRow,
        installed_agents: List[str],
    ) -> Image:
        is_superadmin = ctx.user["role"] == UserRole.SUPERADMIN
        hide_agents = False if is_superadmin else ctx.config_provider.config.manager.hide_agents
        image_ref = row.image_ref
        version, ptag_set = image_ref.tag_set
        ret = cls(
            id=row.id,
            name=row.image,
            namespace=row.image,
            base_image_name=image_ref.name,
            project=row.project,
            humanized_name=row.image,
            tag=row.tag,
            tags=[KVPair(key=k, value=v) for k, v in ptag_set.items()],
            version=version,
            registry=row.registry,
            architecture=row.architecture,
            is_local=row.is_local,
            digest=row.trimmed_digest or None,
            labels=[KVPair(key=k, value=v) for k, v in row.labels.items()],
            aliases=[alias_row.alias for alias_row in row.aliases],
            size_bytes=row.size_bytes,
            status=row.status,
            resource_limits=[
                ResourceLimit(key=k, min=v.get("min", Decimal(0)), max=Decimal("Infinity"))
                for k, v in row.resources.items()
            ],
            supported_accelerators=row.accelerators.split(",") if row.accelerators else ["*"],
            installed=len(installed_agents) > 0,
            installed_agents=installed_agents if not hide_agents else None,
            # legacy
            hash=row.trimmed_digest or None,
        )
        ret.raw_labels = row.labels
        return ret

    @classmethod
    async def from_row(
        cls,
        ctx: GraphQueryContext,
        row: ImageRow,
    ) -> Image:
        # TODO: add architecture
        _installed_agents = await ctx.valkey_image.get_agents_for_image(row.name)
        installed_agents: List[str] = list(_installed_agents)
        return cls.populate_row(ctx, row, installed_agents)

    @classmethod
    async def bulk_load(
        cls,
        ctx: GraphQueryContext,
        rows: List[ImageRow],
    ) -> AsyncIterator[Image]:
        image_canonicals = [row.name for row in rows]
        results = await ctx.valkey_image.get_agents_for_images(image_canonicals)

        for idx, row in enumerate(rows):
            installed_agents: List[str] = list(results[idx])
            yield cls.populate_row(ctx, row, installed_agents)

    @classmethod
    async def batch_load_by_canonical(
        cls,
        graph_ctx: GraphQueryContext,
        image_names: Sequence[str],
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Sequence[Optional[Image]]:
        query = (
            sa.select(ImageRow)
            .where(ImageRow.name.in_(image_names))
            .options(selectinload(ImageRow.aliases))
        )
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))
        async with graph_ctx.db.begin_readonly_session() as session:
            result = await session.execute(query)
            return [await Image.from_row(graph_ctx, row) for row in result.scalars().all()]

    @classmethod
    async def batch_load_by_image_ref(
        cls,
        graph_ctx: GraphQueryContext,
        image_refs: Sequence[ImageRef],
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Sequence[Optional[Image]]:
        image_names = [x.canonical for x in image_refs]
        return await cls.batch_load_by_canonical(graph_ctx, image_names, filter_by_statuses)

    @classmethod
    async def load_item_by_id(
        cls,
        ctx: GraphQueryContext,
        id: UUID,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Image:
        async with ctx.db.begin_readonly_session() as session:
            row = await ImageRow.get(
                session, id, load_aliases=True, filter_by_statuses=filter_by_statuses
            )
            if not row:
                raise ImageNotFound

            return await cls.from_row(ctx, row)

    @classmethod
    async def load_item(
        cls,
        ctx: GraphQueryContext,
        reference: str,
        architecture: str,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Image:
        try:
            async with ctx.db.begin_readonly_session() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageIdentifier(reference, architecture),
                        ImageAlias(reference),
                    ],
                    filter_by_statuses=filter_by_statuses,
                )
        except UnknownImageReference:
            raise ImageNotFound
        return await cls.from_row(ctx, image_row)

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        types: set[ImageLoadFilter] = set(),
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Sequence[Image]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await ImageRow.list(
                session, load_aliases=True, filter_by_statuses=filter_by_statuses
            )
        items: list[Image] = [
            item async for item in cls.bulk_load(ctx, rows) if item.matches_filter(ctx, types)
        ]

        return items

    @staticmethod
    async def filter_allowed(
        ctx: GraphQueryContext,
        items: Sequence[Image],
        domain_name: str,
    ) -> Sequence[Image]:
        from ..domain import domains

        async with ctx.db.begin() as conn:
            query = (
                sa.select([domains.c.allowed_docker_registries])
                .select_from(domains)
                .where(domains.c.name == domain_name)
            )
            result = await conn.execute(query)
            allowed_docker_registries = result.scalar()

        filtered_items: list[Image] = [
            item for item in items if item.registry in allowed_docker_registries
        ]

        return filtered_items

    def matches_filter(
        self,
        ctx: GraphQueryContext,
        load_filters: set[ImageLoadFilter],
    ) -> bool:
        """
        Determine if the image is filtered according to the `load_filters` parameter.
        """
        user_role = ctx.user["role"]

        # If the image filtered by any of its labels, return False early.
        # If the image is not filtered and is determiend to be valid by any of its labels, `is_valid = True`.
        is_valid = ImageLoadFilter.GENERAL in load_filters
        for label in self.labels:
            match label.key:
                case LabelName.FEATURES if KernelFeatures.OPERATION.value in label.value:
                    if ImageLoadFilter.OPERATIONAL in load_filters:
                        is_valid = True
                    else:
                        return False
                case LabelName.CUSTOMIZED_OWNER:
                    if (
                        ImageLoadFilter.CUSTOMIZED not in load_filters
                        and ImageLoadFilter.CUSTOMIZED_GLOBAL not in load_filters
                    ):
                        return False
                    if ImageLoadFilter.CUSTOMIZED in load_filters:
                        if label.value == f"user:{ctx.user['uuid']}":
                            is_valid = True
                        else:
                            return False
                    if ImageLoadFilter.CUSTOMIZED_GLOBAL in load_filters:
                        if user_role == UserRole.SUPERADMIN:
                            is_valid = True
                        else:
                            return False
        return is_valid


class ImagePermissionValueField(graphene.Scalar):
    class Meta:
        description = f"Added in 25.3.0. One of {[val.value for val in ImagePermission]}."

    @staticmethod
    def serialize(val: ImagePermission) -> str:
        return val.value

    @staticmethod
    def parse_literal(node: Any, _variables=None):
        if isinstance(node, graphql.language.ast.StringValueNode):
            return ImagePermission(node.value)

    @staticmethod
    def parse_value(value: str) -> ImagePermission:
        return ImagePermission(value)


@graphene_federation.key("id")
class ImageNode(graphene.ObjectType):
    class Meta:
        interfaces = (AsyncNode,)

    row_id = graphene.UUID(description="Added in 24.03.4. The undecoded id value stored in DB.")
    name = graphene.String(deprecation_reason="Deprecated since 24.12.0. use `namespace` instead")
    namespace = graphene.String(description="Added in 24.12.0.")
    base_image_name = graphene.String(description="Added in 24.12.0.")
    project = graphene.String(description="Added in 24.03.10.")
    humanized_name = graphene.String()
    tag = graphene.String()
    tags = graphene.List(KVPair, description="Added in 24.12.0.")
    version = graphene.String(description="Added in 24.12.0.")
    registry = graphene.String()
    architecture = graphene.String()
    is_local = graphene.Boolean()
    digest = graphene.String()
    labels = graphene.List(KVPair)
    size_bytes = BigInt()
    status = graphene.String(description="Added in 25.4.0.")
    resource_limits = graphene.List(ResourceLimit)
    supported_accelerators = graphene.List(graphene.String)
    aliases = graphene.List(
        graphene.String, description="Added in 24.03.4. The array of image aliases."
    )

    permissions = graphene.List(
        ImagePermissionValueField,
        description=f"Added in 25.3.0. One of {[val.value for val in ImagePermission]}.",
    )
    installed = graphene.Boolean(
        description="Added in 25.11.0. Indicates if the image is installed on any Agent."
    )
    type = ImageTypeEnum(description="Added in 25.12.0.")

    @property
    def _canonical(self) -> str:
        image_ref = ImageRef(
            self.base_image_name,
            self.project,
            self.tag,
            self.registry,
            self.architecture,
            self.is_local,
        )
        return image_ref.canonical

    @classmethod
    async def _batch_load_installed_agents(
        cls, ctx: GraphQueryContext, full_names: Sequence[str]
    ) -> list[set[AgentId]]:
        results = await ctx.valkey_image.get_agents_for_images(list(full_names))
        return [{AgentId(agent_id) for agent_id in agents} for agents in results]

    async def resolve_installed(self, info: graphene.ResolveInfo) -> bool:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, self._batch_load_installed_agents
        )
        agent_ids = await loader.load(self._canonical)
        agent_ids = cast(Optional[set[AgentId]], agent_ids)
        return agent_ids is not None and len(agent_ids) > 0

    @classmethod
    async def batch_load_by_name_and_arch(
        cls,
        graph_ctx: GraphQueryContext,
        name_and_arch: Sequence[tuple[str, str]],
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Sequence[Sequence[ImageNode]]:
        query = (
            sa.select(ImageRow)
            .where(sa.tuple_(ImageRow.name, ImageRow.architecture).in_(name_and_arch))
            .options(selectinload(ImageRow.aliases))
        )
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        async with graph_ctx.db.begin_readonly_session() as db_session:
            return await batch_multiresult_in_scalar_stream(
                graph_ctx,
                db_session,
                query,
                cls,
                name_and_arch,
                lambda row: (row.name, row.architecture),
            )

    @classmethod
    async def batch_load_by_image_identifier(
        cls,
        graph_ctx: GraphQueryContext,
        image_ids: Sequence[ImageIdentifier],
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
    ) -> Sequence[Sequence[ImageNode]]:
        name_and_arch_tuples = [(img.canonical, img.architecture) for img in image_ids]
        return await cls.batch_load_by_name_and_arch(
            graph_ctx, name_and_arch_tuples, filter_by_statuses
        )

    @overload
    @classmethod
    def from_row(cls, graph_ctx: GraphQueryContext, row: ImageRow) -> Self: ...

    @overload
    @classmethod
    def from_row(
        cls, graph_ctx, row: ImageRow, *, permissions: Optional[Iterable[ImagePermission]] = None
    ) -> ImageNode: ...

    @overload
    @classmethod
    def from_row(
        cls, graph_ctx, row: None, *, permissions: Optional[Iterable[ImagePermission]] = None
    ) -> None: ...

    @classmethod
    def from_row(
        cls,
        graph_ctx,
        row: Optional[ImageRow],
        *,
        permissions: Optional[Iterable[ImagePermission]] = None,
    ) -> ImageNode | None:
        if row is None:
            return None
        image_ref = row.image_ref
        version, ptag_set = image_ref.tag_set
        image_type = row.labels.get(LabelName.ROLE, ImageType.COMPUTE.value)

        result = cls(
            id=row.id,
            row_id=row.id,
            name=row.image,
            namespace=row.image,
            base_image_name=image_ref.name,
            project=row.project,
            humanized_name=row.image,
            tag=row.tag,
            tags=[KVPair(key=k, value=v) for k, v in ptag_set.items()],
            version=version,
            registry=row.registry,
            architecture=row.architecture,
            is_local=row.is_local,
            digest=row.trimmed_digest or None,
            labels=[KVPair(key=k, value=v) for k, v in row.labels.items()],
            size_bytes=row.size_bytes,
            resource_limits=[
                ResourceLimit(key=k, min=v.get("min", Decimal(0)), max=Decimal("Infinity"))
                for k, v in row.resources.items()
            ],
            supported_accelerators=row.accelerators.split(",") if row.accelerators else ["*"],
            aliases=[alias_row.alias for alias_row in row.aliases],
            permissions=[] if permissions is None else permissions,
            status=row.status,
            type=image_type,
        )

        return result

    @classmethod
    def from_legacy_image(
        cls, image: Image, *, permissions: Optional[Iterable[ImagePermission]] = None
    ) -> ImageNode:
        labels: dict[str, str] = {kvpair.key: kvpair.value for kvpair in cast(list, image.labels)}
        image_type = labels.get(LabelName.ROLE, ImageType.COMPUTE.value)
        result = cls(
            id=image.id,
            row_id=image.id,
            name=image.name,
            namespace=image.namespace,
            base_image_name=image.base_image_name,
            humanized_name=image.humanized_name,
            tag=image.tag,
            tags=image.tags,
            version=image.version,
            project=image.project,
            registry=image.registry,
            architecture=image.architecture,
            is_local=image.is_local,
            digest=image.digest,
            labels=image.labels,
            size_bytes=image.size_bytes,
            resource_limits=image.resource_limits,
            supported_accelerators=image.supported_accelerators,
            aliases=image.aliases,
            permissions=[] if permissions is None else permissions,
            status=image.status,
            type=image_type,
        )
        return result

    @classmethod
    async def get_node(
        cls,
        info: graphene.ResolveInfo,
        id: ResolvedGlobalID,
        scope_id: ScopeType,
        permission: ImagePermission = ImagePermission.READ_ATTRIBUTE,
    ) -> Optional[Self]:
        graph_ctx: GraphQueryContext = info.context

        _, image_id = id
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return None

            query = (
                sa.select(ImageRow)
                .where(cond & (ImageRow.id == UUID(image_id)))
                .options(selectinload(ImageRow.aliases))
            )

            async with graph_ctx.db.begin_readonly_session() as db_session:
                image_row = await db_session.scalar(query)
                if image_row is None:
                    return None

                return cls.from_row(
                    graph_ctx,
                    image_row,
                    permissions=await permission_ctx.calculate_final_permission(image_row),
                )

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
        permission: ImagePermission,
        filter_by_statuses: Optional[list[ImageStatus]] = [ImageStatus.ALIVE],
        filter_expr: Optional[str] = None,
        order_expr: Optional[str] = None,
        offset: Optional[int] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        before: Optional[str] = None,
        last: Optional[int] = None,
    ) -> ConnectionResolverResult[Self]:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(_queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(_queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            ImageRow,
            ImageRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return ConnectionResolverResult([], cursor, pagination_order, page_size, 0)
            query = query.where(cond).options(selectinload(ImageRow.aliases))
            cnt_query = cnt_query.where(cond)

            if filter_by_statuses:
                query = query.where(ImageRow.status.in_(filter_by_statuses))
                cnt_query = cnt_query.where(ImageRow.status.in_(filter_by_statuses))

            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                image_rows = (await db_session.scalars(query)).all()
                total_cnt = await db_session.scalar(cnt_query)
                result: list[Self] = [
                    cls.from_row(
                        graph_ctx,
                        row,
                        permissions=await permission_ctx.calculate_final_permission(row),
                    )
                    for row in image_rows
                ]
        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)


class ImageConnection(Connection):
    class Meta:
        node = ImageNode
        description = "Added in 25.3.0."


class ForgetImageById(graphene.Mutation):
    """Added in 24.03.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
    ) -> ForgetImageById:
        log.info("forget image {0} by API request", image_id)
        image_uuid = extract_object_uuid(info, image_id, "image")

        ctx: GraphQueryContext = info.context

        result = await ctx.processors.image.forget_image_by_id.wait_for_complete(
            ForgetImageByIdAction(
                user_id=ctx.user["uuid"],
                client_role=ctx.user["role"],
                image_id=image_uuid,
            )
        )

        return ForgetImageById(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class ForgetImage(graphene.Mutation):
    """
    Deprecated since 25.4.0. Use `forget_image_by_id` instead.
    """

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        reference = graphene.String(required=True)
        architecture = graphene.String(default_value=DEFAULT_IMAGE_ARCH)

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        reference: str,
        architecture: str,
    ) -> ForgetImage:
        log.info("forget image {0} by API request", reference)
        ctx: GraphQueryContext = info.context

        result = await ctx.processors.image.forget_image.wait_for_complete(
            ForgetImageAction(
                user_id=ctx.user["uuid"],
                client_role=ctx.user["role"],
                reference=reference,
                architecture=architecture,
            )
        )

        return ForgetImage(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class PurgeImageOptions(graphene.InputObjectType):
    """
    Added in 25.10.0.
    """

    remove_from_registry = graphene.Boolean(
        default_value=False,
        description="Untag the deleted image from the registry. Only available in the HarborV2 registry.",
    )


class PurgeImageById(graphene.Mutation):
    """Added in 25.4.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)
        options = PurgeImageOptions(
            required=False,
            default_value={"remove_from_registry": False},
            description="Added in 25.10.0.",
        )

    image = graphene.Field(ImageNode)

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
        options: PurgeImageOptions,
    ) -> PurgeImageById:
        log.info("purge image row {0} by API request", image_id)
        image_uuid = extract_object_uuid(info, image_id, "image")

        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.purge_image_by_id.wait_for_complete(
            PurgeImageByIdAction(
                user_id=ctx.user["uuid"],
                client_role=ctx.user["role"],
                image_id=image_uuid,
            )
        )

        if options.remove_from_registry:
            await ctx.processors.image.untag_image_from_registry.wait_for_complete(
                UntagImageFromRegistryAction(
                    user_id=ctx.user["uuid"],
                    client_role=ctx.user["role"],
                    image_id=image_uuid,
                )
            )

        return PurgeImageById(image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image)))


class UntagImageFromRegistry(graphene.Mutation):
    """Deprecated since 25.10.0. Use `purge_image_by_id` with `remove_from_registry` option instead."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
    ) -> UntagImageFromRegistry:
        image_uuid = extract_object_uuid(info, image_id, "image")

        log.info("remove image from registry {0} by API request", str(image_uuid))
        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.untag_image_from_registry.wait_for_complete(
            UntagImageFromRegistryAction(
                user_id=ctx.user["uuid"],
                client_role=ctx.user["role"],
                image_id=image_uuid,
            )
        )

        return UntagImageFromRegistry(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class PreloadImage(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        references = graphene.List(graphene.String, required=True)
        target_agents = graphene.List(graphene.String, required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        references: Sequence[str],
        target_agents: Sequence[str],
    ) -> PreloadImage:
        return PreloadImage(ok=False, msg="Not implemented.", task_id=None)


class UnloadImage(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        references = graphene.List(graphene.String, required=True)
        target_agents = graphene.List(graphene.String, required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        references: Sequence[str],
        target_agents: Sequence[str],
    ) -> UnloadImage:
        return UnloadImage(ok=False, msg="Not implemented.", task_id=None)


class RescanImages(graphene.Mutation):
    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        registry = graphene.String()
        project = graphene.String(description="Added in 25.1.0.")

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.UUID()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        registry: Optional[str] = None,
        project: Optional[str] = None,
    ) -> RescanImages:
        log.info(
            "rescanning docker registry {0} by API request",
            f"(registry: {registry or 'all'}, project: {project or 'all'})",
        )
        ctx: GraphQueryContext = info.context

        async def _bg_task(reporter: ProgressReporter) -> DispatchResult:
            loaded_registries: list[ContainerRegistryData]

            if registry is None:
                all_registries = await ctx.processors.container_registry.load_all_container_registries.wait_for_complete(
                    LoadAllContainerRegistriesAction()
                )
                loaded_registries = all_registries.registries
            else:
                registries = await ctx.processors.container_registry.load_container_registries.wait_for_complete(
                    LoadContainerRegistriesAction(
                        registry=registry,
                        project=project,
                    )
                )
                loaded_registries = registries.registries

            rescanned_images = []
            errors = []
            for registry_data in loaded_registries:
                action_result = (
                    await ctx.processors.container_registry.rescan_images.wait_for_complete(
                        RescanImagesAction(
                            registry=registry_data.registry_name,
                            project=registry_data.project,
                            progress_reporter=reporter,
                        )
                    )
                )

                for error in action_result.errors:
                    log.error(error)

                errors.extend(action_result.errors)
                rescanned_images.extend(action_result.images)

            rescanned_image_ids = [image.id for image in rescanned_images]
            if errors:
                return DispatchResult.partial_success(rescanned_image_ids, errors)
            return DispatchResult.success(rescanned_image_ids)

        task_id = await ctx.background_task_manager.start(_bg_task)
        return RescanImages(ok=True, msg="", task_id=task_id)


class AliasImage(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        alias = graphene.String(required=True)
        target = graphene.String(required=True)
        architecture = graphene.String(default_value=DEFAULT_IMAGE_ARCH)

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        alias: str,
        target: str,
        architecture: str,
    ) -> AliasImage:
        log.info("alias image {0} -> {1} by API request", alias, target)
        ctx: GraphQueryContext = info.context

        await ctx.processors.image.alias_image.wait_for_complete(
            AliasImageAction(
                image_canonical=target,
                architecture=architecture,
                alias=alias,
            )
        )

        return AliasImage(ok=True, msg="")


class DealiasImage(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        alias = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        alias: str,
    ) -> DealiasImage:
        log.info("dealias image {0} by API request", alias)
        ctx: GraphQueryContext = info.context

        await ctx.processors.image.dealias_image.wait_for_complete(
            DealiasImageAction(
                alias=alias,
            )
        )

        return DealiasImage(ok=True, msg="")


class ClearImages(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        registry = graphene.String()

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        registry: str,
    ) -> ClearImages:
        ctx: GraphQueryContext = info.context
        log.info("clear images from registry {0} by API request", registry)

        result = (
            await ctx.processors.container_registry.load_container_registries.wait_for_complete(
                LoadContainerRegistriesAction(
                    registry=registry,
                    project=None,
                )
            )
        )

        for registry_data in result.registries:
            await ctx.processors.container_registry.clear_images.wait_for_complete(
                ClearImagesAction(
                    registry=registry_data.registry_name,
                    project=registry_data.project,
                )
            )

        return ClearImages(ok=True, msg="")


class ModifyImageInput(graphene.InputObjectType):
    name = graphene.String(required=False)
    registry = graphene.String(required=False)
    image = graphene.String(required=False)
    tag = graphene.String(required=False)
    architecture = graphene.String(required=False)
    is_local = graphene.Boolean(required=False)
    size_bytes = graphene.Int(required=False)
    type = graphene.String(required=False)
    digest = graphene.String(required=False)
    labels = graphene.List(lambda: KVPairInput, required=False)
    supported_accelerators = graphene.List(graphene.String, required=False)
    resource_limits = graphene.List(lambda: ResourceLimitInput, required=False)

    def to_modifier(self) -> ImageModifier:
        resources_data: dict[str, Any] | UndefinedType = Undefined
        if self.resource_limits is not Undefined:
            resources_data = {}
            for limit_option in self.resource_limits:
                limit_data = {}
                if limit_option.min is not Undefined and len(limit_option.min) > 0:
                    limit_data["min"] = limit_option.min
                if limit_option.max is not Undefined and len(limit_option.max) > 0:
                    limit_data["max"] = limit_option.max
                resources_data[limit_option.key] = limit_data

        accelerators = (
            ",".join(self.supported_accelerators) if self.supported_accelerators else Undefined
        )
        labels = {label.key: label.value for label in self.labels} if self.labels else Undefined

        return ImageModifier(
            name=OptionalState[str].from_graphql(self.name),
            registry=OptionalState[str].from_graphql(self.registry),
            image=OptionalState[str].from_graphql(self.image),
            tag=OptionalState[str].from_graphql(self.tag),
            architecture=OptionalState[str].from_graphql(self.architecture),
            is_local=OptionalState[bool].from_graphql(self.is_local),
            size_bytes=OptionalState[int].from_graphql(self.size_bytes),
            type=OptionalState[ImageType].from_graphql(self.type),
            config_digest=OptionalState[str].from_graphql(self.digest),
            labels=OptionalState[dict[str, Any]].from_graphql(labels),
            accelerators=TriState[str].from_graphql(
                accelerators,
            ),
            resources=OptionalState[dict[str, Any]].from_graphql(
                resources_data,
            ),
        )


class ModifyImage(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        target = graphene.String(required=True, default_value=None)
        architecture = graphene.String(required=False, default_value=DEFAULT_IMAGE_ARCH)
        props = ModifyImageInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        target: str,
        architecture: str,
        props: ModifyImageInput,
    ) -> AliasImage:
        ctx: GraphQueryContext = info.context
        log.info("modify image {0} by API request", target)

        await ctx.processors.image.modify_image.wait_for_complete(
            ModifyImageAction(
                target=target,
                architecture=architecture,
                modifier=props.to_modifier(),
            )
        )

        return ModifyImage(ok=True, msg="")


class PurgeImagesKey(graphene.InputObjectType):
    """
    Added in 25.6.0.
    """

    agent_id = graphene.String(required=True)
    images = graphene.List(ImageRefType, required=True)


class PurgeImagesOptions(graphene.InputObjectType):
    """
    Added in 25.6.0.
    """

    force = graphene.Boolean(
        default_value=False,
        description="Remove the images even if it is being used by stopped containers or has other tags, Added in 25.6.0.",
    )
    noprune = graphene.Boolean(
        default_value=False, description="Don't delete untagged parent images, Added in 25.6.0."
    )


class PurgeImagesPayload(graphene.ObjectType):
    """
    Added in 25.6.0.
    """

    task_id = graphene.String()
    allowed_roles = (UserRole.SUPERADMIN, UserRole.ADMIN)


class PurgeImages(graphene.Mutation):
    """
    Added in 25.4.0.
    """

    class Arguments:
        keys = graphene.List(PurgeImagesKey, required=True)
        options = PurgeImagesOptions(default_value={"force": False, "noprune": False})

    Output = PurgeImagesPayload

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        keys: list[PurgeImagesKey],
        options: PurgeImagesOptions,
    ) -> PurgeImagesPayload:
        ctx: GraphQueryContext = info.context
        agent_images = ", ".join(
            f"{key.agent_id}: [{', '.join(img.name for img in key.images)}]" for key in keys
        )

        log.info(f"purge images ({agent_images}) by API request")

        async def _bg_task(reporter: ProgressReporter) -> DispatchResult:
            total_result: PurgeImagesActionResult = PurgeImagesActionResult(
                total_reserved_bytes=0,
                purged_images=[],
                errors=[],
            )

            for key in keys:
                agent_id = key.agent_id
                for img in key.images:
                    # TODO: Use asyncio.gather?
                    result: PurgeImageActionResult = (
                        await ctx.processors.image.purge_image.wait_for_complete(
                            PurgeImageAction(
                                ImageRefData(
                                    name=img.name,
                                    registry=img.registry,
                                    architecture=img.architecture,
                                ),
                                agent_id=agent_id,
                                force=options.force,
                                noprune=options.noprune,
                            )
                        )
                    )

                    total_result.total_reserved_bytes += result.reserved_bytes
                    total_result.purged_images.append(
                        PurgedImagesData(
                            agent_id=agent_id,
                            purged_images=[result.purged_image.name],
                        )
                    )

                    if result.error is not None:
                        log.error(result.error)
                        total_result.errors.append(result.error)

            if total_result.errors:
                return DispatchResult.partial_success(
                    total_result.purged_images, total_result.errors
                )
            return DispatchResult.success(total_result.purged_images)

        task_id = await ctx.background_task_manager.start(_bg_task)
        return PurgeImagesPayload(task_id=task_id)


class ClearImageCustomResourceLimitKey(graphene.InputObjectType):
    """
    Added in 25.6.0.
    """

    image_canonical = graphene.String(required=True)
    architecture = graphene.String(required=True, default_value=DEFAULT_IMAGE_ARCH)


class ClearImageCustomResourceLimitPayload(graphene.ObjectType):
    """
    Added in 25.6.0.
    """

    image_node = graphene.Field(ImageNode)
    allowed_roles = (UserRole.SUPERADMIN,)


class ClearImageCustomResourceLimit(graphene.Mutation):
    """
    Added in 25.6.0.
    """

    class Arguments:
        key = ClearImageCustomResourceLimitKey(required=True)

    Output = ClearImageCustomResourceLimitPayload

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        key: ClearImageCustomResourceLimitKey,
    ) -> ClearImageCustomResourceLimitPayload:
        log.info(
            f'clear custom resource limits for image "{key.image_canonical}" ({key.architecture}) by API request',
        )
        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.clear_image_custom_resource_limit.wait_for_complete(
            ClearImageCustomResourceLimitAction(
                image_canonical=key.image_canonical,
                architecture=key.architecture,
            )
        )
        return ClearImageCustomResourceLimitPayload(
            image_node=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image_data))
        )



import base64
from datetime import datetime, timedelta
from enum import Enum
from typing import Annotated, Any, Optional, cast
from uuid import uuid4

import strawberry
from strawberry import ID, Info, relay
from strawberry.relay import Connection, Edge, PageInfo
from strawberry.relay.types import NodeIterableType

from ai.backend.manager.api.gql.base import JSONString, OrderDirection, StringFilter
from ai.backend.manager.api.gql.federated_types import Image, VFolder


@strawberry.type
class ModelMountConfig:
    vfolder: VFolder
    mount_destination: str
    definition_path: str


@strawberry.type
class RawServiceConfig:
    config: JSONString
    extra_cli_parameters: Optional[str] = None


ServiceConfig = Annotated[
    RawServiceConfig,
    strawberry.union(
        "ServiceConfig", description="Different service configurations for model runtime"
    ),
]


@strawberry.type
class ModelRuntimeConfig:
    runtime_variant: str
    service_config: Optional[ServiceConfig] = None
    environ: Optional[JSONString] = None


@strawberry.type
class ModelRevision(relay.Node):
    id: relay.NodeID
    name: str

    model_runtime_config: ModelRuntimeConfig
    model_mount_config: ModelMountConfig

    image: Image

    created_at: datetime


# Filter and Order Types
@strawberry.input
class ModelRevisionFilter:
    name: Optional[StringFilter] = None
    deployment_id: Optional[ID] = None

    AND: Optional["ModelRevisionFilter"] = None
    OR: Optional["ModelRevisionFilter"] = None
    NOT: Optional["ModelRevisionFilter"] = None
    DISTINCT: Optional[bool] = None


@strawberry.enum
class ModelRevisionOrderField(Enum):
    CREATED_AT = "CREATED_AT"
    NAME = "NAME"


@strawberry.input
class ModelRevisionOrder:
    field: ModelRevisionOrderField
    direction: OrderDirection = OrderDirection.DESC


# TODO: After implementing the actual logic, remove these mock objects
# Mock Model Revisions
def _generate_mock_global_id() -> str:
    return base64.b64encode(f"default:{uuid4()}".encode("utf-8")).decode()


mock_model_revision_1 = ModelRevision(
    id=_generate_mock_global_id(),
    name="llama-3-8b-instruct-v1.0",
    model_runtime_config=ModelRuntimeConfig(
        runtime_variant="vllm",
        service_config=RawServiceConfig(
            config=cast(
                JSONString,
                '{"max_model_length": 4096, "parallelism": {"tensor_parallel_size": 1}, "extra_cli_parameters": "--enable-prefix-caching"}',
            ),
        ),
        environ=cast(JSONString, '{"CUDA_VISIBLE_DEVICES": "0"}'),
    ),
    model_mount_config=ModelMountConfig(
        vfolder=VFolder(id=ID(_generate_mock_global_id())),
        mount_destination="/models",
        definition_path="models/llama-3-8b/config.yaml",
    ),
    image=Image(id=ID(_generate_mock_global_id())),
    created_at=datetime.now() - timedelta(days=10),
)

mock_model_revision_2 = ModelRevision(
    id=_generate_mock_global_id(),
    name="llama-3-8b-instruct-v1.1",
    model_runtime_config=ModelRuntimeConfig(
        runtime_variant="vllm",
        service_config=RawServiceConfig(
            config=cast(
                JSONString,
                '{"max_model_length": 4096, "parallelism": {"tensor_parallel_size": 1}, "extra_cli_parameters": "--enable-prefix-caching"}',
            ),
        ),
        environ=cast(JSONString, '{"CUDA_VISIBLE_DEVICES": "0,1"}'),
    ),
    model_mount_config=ModelMountConfig(
        vfolder=VFolder(id=ID(_generate_mock_global_id())),
        mount_destination="/models",
        definition_path="models/llama-3-8b/config.yaml",
    ),
    image=Image(id=ID(_generate_mock_global_id())),
    created_at=datetime.now() - timedelta(days=5),
)

mock_model_revision_3 = ModelRevision(
    id=_generate_mock_global_id(),
    name="mistral-7b-v0.3-initial",
    model_runtime_config=ModelRuntimeConfig(
        runtime_variant="vllm",
        service_config=RawServiceConfig(
            config=cast(
                JSONString,
                '{"max_model_length": 4096, "parallelism": {"tensor_parallel_size": 1}, "extra_cli_parameters": "--enable-prefix-caching"}',
            ),
        ),
        environ=cast(JSONString, '{"CUDA_VISIBLE_DEVICES": "2"}'),
    ),
    model_mount_config=ModelMountConfig(
        vfolder=VFolder(id=ID(_generate_mock_global_id())),
        mount_destination="/models",
        definition_path="models/mistral-7b/config.yaml",
    ),
    image=Image(id=ID(_generate_mock_global_id())),
    created_at=datetime.now() - timedelta(days=20),
)


# Payload Types
@strawberry.type
class CreateModelRevisionPayload:
    revision: ModelRevision


# Input Types
@strawberry.input
class ImageInput:
    name: str
    architecture: str


@strawberry.input
class ModelRuntimeConfigInput:
    runtime_variant: str
    service_config: Optional[JSONString] = None
    environ: Optional[JSONString] = None


@strawberry.input
class ModelMountConfigInput:
    vfolder_id: ID
    mount_destination: str
    definition_path: str


@strawberry.input
class CreateModelRevisionInput:
    deployment_id: ID
    name: str
    image: ImageInput
    model_runtime_config: ModelRuntimeConfigInput
    model_mount_config: ModelMountConfigInput


ModelRevisionEdge = Edge[ModelRevision]


@strawberry.type
class ModelRevisionConnection(Connection[ModelRevision]):
    """Connection type for ModelRevision, used for Relay pagination."""

    @strawberry.field
    def count(self) -> int:
        return 0

    @classmethod
    def resolve_connection(
        cls,
        nodes: NodeIterableType[ModelRevision],
        *,
        info: Optional[Info] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        max_results: Optional[int] = None,
        **kwargs: Any,
    ):
        """Resolve the connection for Relay pagination."""
        revisions = [mock_model_revision_1, mock_model_revision_2, mock_model_revision_3]
        edges = [ModelRevisionEdge(node=rev, cursor=str(i)) for i, rev in enumerate(revisions)]
        return cls(
            edges=edges,
            page_info=PageInfo(
                has_next_page=False, has_previous_page=False, start_cursor=None, end_cursor=None
            ),
        )


@strawberry.relay.connection(ModelRevisionConnection)
async def revisions(
    filter: Optional[ModelRevisionFilter] = None,
    order: Optional[ModelRevisionOrder] = None,
    first: Optional[int] = None,
    after: Optional[str] = None,
) -> list[ModelRevision]:
    """List revisions with optional filtering and pagination."""
    return [mock_model_revision_1, mock_model_revision_2, mock_model_revision_3]


@strawberry.field
async def revision(id: ID) -> Optional[ModelRevision]:
    """Get a specific revision by ID."""
    return None


@strawberry.mutation
async def create_model_revision(input: CreateModelRevisionInput) -> CreateModelRevisionPayload:
    """Create a new model revision."""
    revision = ModelRevision(
        id=_generate_mock_global_id(),
        name=input.name,
        model_runtime_config=ModelRuntimeConfig(
            runtime_variant=input.model_runtime_config.runtime_variant,
            service_config=None,
            environ=None,
        ),
        model_mount_config=ModelMountConfig(
            vfolder=VFolder(id=ID(_generate_mock_global_id())),
            mount_destination="/models",
            definition_path="model.yaml",
        ),
        image=Image(id=ID(_generate_mock_global_id())),
        created_at=datetime.now(),
    )
    return CreateModelRevisionPayload(revision=revision)



"""
Federated types that reference existing GraphQL entities.
These types connect the new Strawberry schema with the existing Graphene schema.
"""

import strawberry
from strawberry import ID


@strawberry.federation.type(name="UserNode", keys=["id"], extend=True)
class User:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID):
        return cls(id=id)


@strawberry.federation.type(name="ScalingGroupNode", keys=["id"], extend=True)
class ResourceGroup:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID):
        return cls(id=id)


@strawberry.federation.type(name="EndpointToken", keys=["token"], extend=True)
class AccessToken:
    token: str = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, token: str) -> "AccessToken":
        return cls(token=token)


@strawberry.federation.type(name="VirtualFolderNode", keys=["id"], extend=True)
class VFolder:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID) -> "VFolder":
        return cls(id=id)


@strawberry.federation.type(name="EndpointAutoScalingRuleNode", keys=["id"], extend=True)
class AutoScalingRule:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID) -> "AutoScalingRule":
        return cls(id=id)


@strawberry.federation.type(name="ImageNode", keys=["id"], extend=True)
class Image:
    id: ID = strawberry.federation.field(external=True)

    @classmethod
    def resolve_reference(cls, id: ID) -> "Image":
        return cls(id=id)



from typing import Any, Mapping

import sqlalchemy as sa

from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.data.resource.types import KeyPairResourcePolicyData
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

# Layer-specific decorator for keypair_resource_policy repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.KEYPAIR_RESOURCE_POLICY)


class KeypairResourcePolicyRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @repository_decorator()
    async def create(self, fields: Mapping[str, Any]) -> KeyPairResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            db_row = KeyPairResourcePolicyRow(**fields)
            db_sess.add(db_row)
            await db_sess.flush()
            # Refresh the object to ensure all attributes are loaded
            await db_sess.refresh(db_row)
            return db_row.to_dataclass()

    @repository_decorator()
    async def get_by_name(self, name: str) -> KeyPairResourcePolicyData:
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(KeyPairResourcePolicyRow).where(KeyPairResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Keypair resource policy with name {name} not found.")
            return row.to_dataclass()

    @repository_decorator()
    async def update(self, name: str, fields: Mapping[str, Any]) -> KeyPairResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            query = sa.select(KeyPairResourcePolicyRow).where(KeyPairResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Keypair resource policy with name {name} not found.")
            for key, value in fields.items():
                setattr(row, key, value)
            await db_sess.flush()
            return row.to_dataclass()

    @repository_decorator()
    async def delete(self, name: str) -> KeyPairResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            query = sa.select(KeyPairResourcePolicyRow).where(KeyPairResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Keypair resource policy with name {name} not found.")
            await db_sess.delete(row)
            return row.to_dataclass()



"""Base classes for session state transition hooks."""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ai.backend.common.types import ClusterMode
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.agent.pool import AgentPool
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.models.network import NetworkType
from ai.backend.manager.plugin.network import NetworkPluginContext

from ..types import SessionTransitionData

log = BraceStyleAdapter(logging.getLogger(__name__))


class AbstractSessionHook(ABC):
    """
    Abstract base class for session state transition hooks.
    Subclasses implement session-type specific logic.
    """

    @abstractmethod
    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        """
        Called when a session is about to transition from CREATING to RUNNING.
        Raises exception if the transition should not proceed.

        :param session: Session transition data with all necessary information
        :raises Exception: If the hook fails and transition should not proceed
        """
        raise NotImplementedError

    @abstractmethod
    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        """
        Called when a session is about to transition from TERMINATING to TERMINATED.
        Best-effort cleanup - exceptions are logged but don't prevent termination.

        :param session: Session transition data with all necessary information
        :raises Exception: If cleanup fails (will be logged but ignored)
        """
        raise NotImplementedError


@dataclass
class SessionHookArgs:
    network_plugin_ctx: NetworkPluginContext
    config_provider: ManagerConfigProvider
    agent_pool: AgentPool


class SessionHook(AbstractSessionHook):
    _session_hook: AbstractSessionHook
    _network_plugin_ctx: NetworkPluginContext
    _config_provider: ManagerConfigProvider
    _agent_pool: AgentPool

    def __init__(self, session_hook: AbstractSessionHook, args: SessionHookArgs) -> None:
        self._session_hook = session_hook
        self._network_plugin_ctx = args.network_plugin_ctx
        self._config_provider = args.config_provider
        self._agent_pool = args.agent_pool

    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        # Bypass to the actual hook implementation
        await self._session_hook.on_transition_to_running(session)

    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        try:
            await self._session_hook.on_transition_to_terminated(session)
        except Exception as e:
            log.error(
                "Error during cleanup for session {}: {}",
                session.session_id,
                str(e),
            )
        finally:
            log.info(
                "Running default cleanup for session transition data {}",
                session,
            )
            if session.network_type == NetworkType.VOLATILE:
                await self._destroy_network(session)

    async def _destroy_network(self, session: SessionTransitionData) -> None:
        log.info(
            "Destroying network {} for session {} ...",
            session.network_id,
            session.session_id,
        )
        network_id = session.network_id
        if network_id is None:
            log.debug(
                "No network to destroy for session {} (network_id is None)",
                session.session_id,
            )
            return
        match session.cluster_mode:
            case ClusterMode.SINGLE_NODE:
                try:
                    agent_client = self._agent_pool.get_agent_client(
                        session.main_kernel.agent_id,
                        order_key=str(session.session_id),
                    )
                    await agent_client.destroy_local_network(network_id)
                except Exception:
                    log.exception(f"Failed to destroy the agent-local network {network_id}")
            case ClusterMode.MULTI_NODE:
                if self._config_provider.config.network.inter_container.default_driver is None:
                    raise ValueError("No inter-container network driver is configured.")

                network_plugin = self._network_plugin_ctx.plugins[
                    self._config_provider.config.network.inter_container.default_driver
                ]
                try:
                    await network_plugin.destroy_network(network_id=network_id)
                except Exception:
                    log.exception(f"Failed to destroy the overlay network {network_id}.")


class NoOpSessionHook(AbstractSessionHook):
    """Default no-op hook for session types that don't need special handling."""

    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        log.debug(
            "No-op hook for session {} (type: {})",
            session.session_id,
            session.session_type,
        )

    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        log.debug(
            "No-op cleanup for session {} (type: {})",
            session.session_id,
            session.session_type,
        )



from collections.abc import Mapping
from contextlib import asynccontextmanager as actxmgr
from dataclasses import dataclass
from http import HTTPStatus
from typing import Any, AsyncIterator, Final, Optional

import aiohttp
import yarl

from ai.backend.common.json import load_json
from ai.backend.manager.errors.storage import (
    UnexpectedStorageProxyResponseError,
    VFolderBadRequest,
    VFolderGone,
    VFolderNotFound,
    VFolderOperationFailed,
)

AUTH_TOKEN_HDR: Final = "X-BackendAI-Storage-Auth-Token"


@dataclass
class StorageProxyClientArgs:
    endpoint: yarl.URL
    secret: str


class StorageProxyHTTPClient:
    _client_session: aiohttp.ClientSession
    _endpoint: yarl.URL
    _secret: str

    def __init__(self, client_session: aiohttp.ClientSession, args: StorageProxyClientArgs):
        self._client_session = client_session
        self._endpoint = args.endpoint
        self._secret = args.secret

    @actxmgr
    async def request_stream_response(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        """
        Make an HTTP request using the session client.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :return: Response data as a dictionary, or None if no content
        """
        headers = {
            AUTH_TOKEN_HDR: self._secret,
        }
        async with self._client_session.request(
            method,
            self._endpoint / url,
            headers=headers,
            json=body,
            params=params,
        ) as client_resp:
            if client_resp.status // 100 == 2:
                yield client_resp
                return
            match client_resp.status:
                case HTTPStatus.BAD_REQUEST:
                    raise VFolderBadRequest(
                        extra_msg="Bad request to storage proxy",
                    )
                case HTTPStatus.NOT_FOUND:
                    raise VFolderNotFound(
                        extra_msg="Requested resource not found",
                    )
                case HTTPStatus.GONE:
                    raise VFolderGone(
                        extra_msg=(
                            "The requested resource is gone. It may have been deleted or moved."
                        ),
                    )
                case HTTPStatus.INTERNAL_SERVER_ERROR:
                    raise VFolderOperationFailed(
                        extra_msg="Internal server error from storage proxy",
                    )
                case _:
                    raise VFolderOperationFailed(
                        extra_msg=f"Unexpected error {client_resp.status} from storage proxy",
                    )

    async def request(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> Optional[Mapping[str, Any]]:
        """
        Make an HTTP request using the session client.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :return: Response data as a dictionary, or None if no content
        """
        async with self.request_stream_response(
            method, url, body=body, params=params
        ) as response_stream:
            if response_stream.status == HTTPStatus.NO_CONTENT:
                return None
            resp_bytes = await response_stream.read()
            if not resp_bytes:
                return None
            return load_json(resp_bytes)

    async def request_with_response(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> Mapping[str, Any]:
        """
        Make an HTTP request and return the response as a dictionary.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :return: Response object from the request
        """
        response = await self.request(method, url, body=body, params=params)
        if response is None:
            raise UnexpectedStorageProxyResponseError(
                "Unexpected response from storage proxy: None",
            )
        return response



"""Handler for checking route health status."""

import logging
from collections.abc import Sequence
from typing import Optional

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.model_serving.types import RouteStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class HealthCheckRouteHandler(RouteHandler):
    """Handler for checking route health status (readiness and liveness)."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ):
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "health-check-routes"

    @property
    def lock_id(self) -> Optional[LockID]:
        """Lock for health check routes."""
        return LockID.LOCKID_DEPLOYMENT_HEALTH_CHECK_ROUTES

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.HEALTHY, RouteStatus.UNHEALTHY]

    @classmethod
    def next_status(cls) -> Optional[RouteStatus]:
        """Routes that pass health check become HEALTHY."""
        return RouteStatus.HEALTHY

    @classmethod
    def failure_status(cls) -> Optional[RouteStatus]:
        """Routes that fail health check become UNHEALTHY."""
        return RouteStatus.UNHEALTHY

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute health check for routes."""
        log.debug("Checking health for {} routes", len(routes))

        # Execute route health check logic via executor
        result = await self._route_executor.check_route_health(routes)
        return result

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after health check."""
        healthy_count = len(result.successes)
        unhealthy_count = len(result.errors)

        if unhealthy_count > 0:
            log.debug(
                "Health check complete: {} healthy, {} unhealthy routes",
                healthy_count,
                unhealthy_count,
            )

            # Log details of unhealthy routes
            for error in result.errors:
                log.trace("Route {} is unhealthy: {}", error.route_info.route_id, error.reason)
        else:
            log.trace("All {} routes are healthy", healthy_count)



from typing import Optional

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.data.domain.types import (
    DomainCreator,
    DomainData,
    DomainModifier,
    UserInfo,
)
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.resource import (
    DomainDataProcessingError,
    DomainDeletionFailed,
    DomainHasActiveKernels,
    DomainHasGroups,
    DomainHasUsers,
    DomainNotFound,
)
from ai.backend.manager.models import groups
from ai.backend.manager.models.domain import DomainRow, domains, row_to_data
from ai.backend.manager.models.group import ProjectType
from ai.backend.manager.models.kernel import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    kernels,
)
from ai.backend.manager.models.scaling_group import ScalingGroupForDomainRow
from ai.backend.manager.models.user import users
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_txn_retry

# Layer-specific decorator for admin domain repository
admin_repository_decorator = create_layer_aware_repository_decorator(LayerType.DOMAIN)


class AdminDomainRepository:
    """
    Repository for admin-specific domain operations that bypass ownership checks.
    This should only be used by superadmin users.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def create_domain_force(self, creator: DomainCreator) -> DomainData:
        """
        Creates a new domain with model-store group without permission checks.
        For superadmin use only.
        """
        async with self._db.begin() as conn:
            check_query = sa.select(DomainRow).where(DomainRow.name == creator.name)
            existing_domain = await conn.scalar(check_query)
            if existing_domain is not None:
                raise InvalidAPIParameters(f"Domain with name '{creator.name}' already exists")

            data = creator.fields_to_store()
            insert_query = sa.insert(domains).values(data).returning(domains)
            result = await conn.execute(insert_query)
            row = result.first()

            # Create model-store group for the domain
            await self._create_model_store_group(conn, creator.name)

            if result.rowcount != 1 or row is None:
                raise RuntimeError(f"No domain created. rowcount: {result.rowcount}, data: {data}")

        assert row is not None
        result = row_to_data(row)
        assert result is not None
        return result

    @admin_repository_decorator()
    async def modify_domain_force(self, domain_name: str, modifier: DomainModifier) -> DomainData:
        """
        Modifies an existing domain without permission checks.
        For superadmin use only.
        """
        async with self._db.begin() as conn:
            data = modifier.fields_to_update()
            update_query = (
                sa.update(domains)
                .values(data)
                .where(domains.c.name == domain_name)
                .returning(domains)
            )
            result = await conn.execute(update_query)
            row = result.first()

            if result.rowcount == 0:
                raise DomainNotFound(f"Domain not found: {domain_name}")
            return row_to_data(row)

    @admin_repository_decorator()
    async def soft_delete_domain_force(self, domain_name: str) -> None:
        """
        Soft deletes a domain by setting is_active to False without permission checks.
        For superadmin use only.
        """
        async with self._db.begin() as conn:
            update_query = (
                sa.update(domains).values({"is_active": False}).where(domains.c.name == domain_name)
            )
            result = await conn.execute(update_query)
            if result.rowcount == 0:
                raise DomainNotFound(f"Domain not found: {domain_name}")

    @admin_repository_decorator()
    async def purge_domain_force(self, domain_name: str) -> None:
        """
        Permanently deletes a domain without validation checks.
        """
        async with self._db.begin() as conn:
            # Validate prerequisites
            if await self._domain_has_active_kernels(conn, domain_name):
                raise DomainHasActiveKernels(
                    "Domain has some active kernels. Terminate them first."
                )

            user_count = await self._get_domain_user_count(conn, domain_name)
            if user_count > 0:
                raise DomainHasUsers("There are users bound to the domain. Remove users first.")

            group_count = await self._get_domain_group_count(conn, domain_name)
            if group_count > 0:
                raise DomainHasGroups("There are groups bound to the domain. Remove groups first.")

            # Clean up kernels
            await self._delete_kernels(conn, domain_name)

            # Delete domain
            delete_query = sa.delete(domains).where(domains.c.name == domain_name)
            result = await conn.execute(delete_query)
            if result.rowcount == 0:
                raise DomainDeletionFailed(f"Failed to delete domain: {domain_name}")

    async def _domain_has_active_kernels(self, conn: SAConnection, domain_name: str) -> bool:
        """
        Private method to check if domain has active kernels.
        """
        query = (
            sa.select([sa.func.count()])
            .select_from(kernels)
            .where(
                (kernels.c.domain_name == domain_name)
                & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))
            )
        )
        active_kernel_count = await conn.scalar(query)
        return active_kernel_count > 0

    async def _get_domain_user_count(self, conn: SAConnection, domain_name: str) -> int:
        """
        Private method to get user count for a domain.
        """
        query = sa.select([sa.func.count()]).where(users.c.domain_name == domain_name)
        return await conn.scalar(query)

    async def _get_domain_group_count(self, conn: SAConnection, domain_name: str) -> int:
        """
        Private method to get group count for a domain.
        """
        query = sa.select([sa.func.count()]).where(groups.c.domain_name == domain_name)
        return await conn.scalar(query)

    async def create_domain_node_force(
        self, creator: DomainCreator, scaling_groups: Optional[list[str]] = None
    ) -> DomainData:
        """
        Creates a domain node with scaling groups without permission checks.
        For superadmin use only.
        """
        async with self._db.begin_session() as session:
            data = creator.fields_to_store()
            check_query = sa.select(DomainRow).where(DomainRow.name == creator.name)
            existing_domain = await session.scalar(check_query)
            if existing_domain is not None:
                raise InvalidAPIParameters(f"Domain with name '{creator.name}' already exists")

            insert_and_returning = sa.select(DomainRow).from_statement(
                sa.insert(DomainRow).values(data).returning(DomainRow)
            )
            domain_row = await session.scalar(insert_and_returning)
            if domain_row is None:
                raise DomainDataProcessingError(f"Failed to create domain node: {creator.name}")

            if scaling_groups is not None:
                await session.execute(
                    sa.insert(ScalingGroupForDomainRow),
                    [
                        {"scaling_group": sgroup_name, "domain": creator.name}
                        for sgroup_name in scaling_groups
                    ],
                )

            await session.commit()
            result = row_to_data(domain_row)
            assert result is not None
            return result

    async def _create_model_store_group(self, conn: SAConnection, domain_name: str) -> None:
        """
        Private method to create model-store group for a domain.
        """
        model_store_insert_query = sa.insert(groups).values({
            "name": "model-store",
            "description": "Model Store",
            "is_active": True,
            "domain_name": domain_name,
            "total_resource_slots": {},
            "allowed_vfolder_hosts": {},
            "integration_id": None,
            "resource_policy": "default",
            "type": ProjectType.MODEL_STORE,
        })
        await conn.execute(model_store_insert_query)

    async def _delete_kernels(self, conn: SAConnection, domain_name: str) -> int:
        """
        Private method to delete all kernels for a domain.
        """
        delete_query = sa.delete(kernels).where(kernels.c.domain_name == domain_name)
        result = await conn.execute(delete_query)
        return result.rowcount

    async def create_domain_node_with_permissions_force(
        self,
        creator: DomainCreator,
        user_info: UserInfo,
        scaling_groups: Optional[list[str]] = None,
    ) -> DomainData:
        """
        Creates a domain node with scaling groups without permission checks.
        For superadmin use only.
        """

        async def _insert(db_session: SASession) -> DomainData:
            return await self.create_domain_node_force(creator, scaling_groups)

        async with self._db.connect() as db_conn:
            return await execute_with_txn_retry(_insert, self._db.begin_session, db_conn)

    async def modify_domain_node_with_permissions_force(
        self,
        domain_name: str,
        modifier_fields: dict,
        user_info: UserInfo,
        sgroups_to_add: Optional[set[str]] = None,
        sgroups_to_remove: Optional[set[str]] = None,
    ) -> DomainData:
        """
        Modifies a domain node with scaling group changes without permission checks.
        For superadmin use only.
        """
        async with self._db.begin_session() as session:
            if sgroups_to_add is not None:
                await session.execute(
                    sa.insert(ScalingGroupForDomainRow),
                    [
                        {"scaling_group": sgroup_name, "domain": domain_name}
                        for sgroup_name in sgroups_to_add
                    ],
                )

            if sgroups_to_remove is not None:
                await session.execute(
                    sa.delete(ScalingGroupForDomainRow).where(
                        (ScalingGroupForDomainRow.domain == domain_name)
                        & (ScalingGroupForDomainRow.scaling_group.in_(sgroups_to_remove))
                    ),
                )

            update_stmt = (
                sa.update(DomainRow)
                .where(DomainRow.name == domain_name)
                .values(modifier_fields)
                .returning(DomainRow)
            )
            await session.execute(update_stmt)

            domain_row = await session.scalar(
                sa.select(DomainRow).where(DomainRow.name == domain_name)
            )

            await session.commit()
            if domain_row is None:
                raise DomainNotFound(f"Domain not found (id:{domain_name})")
            return row_to_data(domain_row)



from .association_scopes_entities import AssociationScopesEntitiesRow
from .permission.object_permission import ObjectPermissionRow
from .permission.permission import PermissionRow
from .permission.permission_group import PermissionGroupRow
from .role import RoleRow
from .user_role import UserRoleRow

__all__ = (
    "AssociationScopesEntitiesRow",
    "ObjectPermissionRow",
    "PermissionRow",
    "PermissionGroupRow",
    "RoleRow",
    "UserRoleRow",
)



from dataclasses import dataclass, field
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.domain.types import DomainData, DomainNodeModifier, UserInfo
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class ModifyDomainNodeAction(DomainAction):
    name: str
    user_info: UserInfo
    sgroups_to_add: Optional[set[str]] = None
    sgroups_to_remove: Optional[set[str]] = None
    modifier: DomainNodeModifier = field(default_factory=DomainNodeModifier)

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "modify"


@dataclass
class ModifyDomainNodeActionResult(BaseActionResult):
    domain_data: DomainData

    @override
    def entity_id(self):
        return self.domain_data.name



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.image.admin_repository import AdminImageRepository
from ai.backend.manager.repositories.image.repository import ImageRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ImageRepositories:
    repository: ImageRepository
    admin_repository: AdminImageRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ImageRepository(args.db)
        admin_repository = AdminImageRepository(args.db)

        return cls(
            repository=repository,
            admin_repository=admin_repository,
        )



"""Types for session creation and enqueueing."""

import secrets
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any, Optional, Self
from uuid import UUID

import yarl

from ai.backend.common.defs.session import SESSION_PRIORITY_DEFAULT
from ai.backend.common.docker import ImageRef
from ai.backend.common.types import (
    AccessKey,
    ClusterMode,
    KernelEnqueueingConfig,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionTypes,
    VFolderMount,
)
from ai.backend.manager.data.deployment.types import DeploymentInfo
from ai.backend.manager.models import NetworkRow
from ai.backend.manager.models.network import NetworkType
from ai.backend.manager.models.scaling_group import ScalingGroupOpts
from ai.backend.manager.types import UserScope


@dataclass
class UserContext:
    """User information for session creation."""

    uuid: UUID
    access_key: AccessKey
    role: str  # UserRole as string
    sudo_session_enabled: bool


@dataclass
class ContainerUserContext:
    """Container user UID/GID information."""

    uid: Optional[int]
    main_gid: Optional[int]
    supplementary_gids: list[int]


@dataclass
class ImageContext:
    """Resolved image information."""

    ref: ImageRef  # Image reference object
    labels: dict[str, Any]


@dataclass
class DeploymentContext:
    """Context data needed to create a session from deployment info."""

    created_user: UserContext
    session_owner: UserContext
    container_user: ContainerUserContext
    group_id: UUID
    resource_policy: dict[str, Any]
    image: ImageContext


@dataclass
class SessionUserInfo:
    user_scope: UserScope
    access_key: AccessKey
    resource_policy: dict[str, Any]  # Raw policy data


@dataclass
class SessionCreationSpec:
    """Specification for creating a new session."""

    # Required parameters
    session_creation_id: str
    session_name: str
    access_key: AccessKey
    user_scope: UserScope
    session_type: SessionTypes
    cluster_mode: ClusterMode
    cluster_size: int
    priority: int
    resource_policy: dict
    kernel_specs: list[KernelEnqueueingConfig]
    creation_spec: dict

    # Optional parameters
    scaling_group: Optional[str] = None
    session_tag: Optional[str] = None
    starts_at: Optional[datetime] = None
    batch_timeout: Optional[timedelta] = None
    dependency_sessions: Optional[list[SessionId]] = None
    callback_url: Optional[yarl.URL] = None
    route_id: Optional[UUID] = None
    sudo_session_enabled: bool = False
    network: Optional[NetworkRow] = None
    designated_agent_list: Optional[list[str]] = None
    internal_data: Optional[dict] = None
    public_sgroup_only: bool = True

    @classmethod
    def from_deployment_info(
        cls, deployment_info: DeploymentInfo, context: DeploymentContext, route_id: UUID
    ) -> Self:
        session_creation_id = secrets.token_urlsafe(16)
        target_revision = deployment_info.target_revision()
        if target_revision is None:
            raise ValueError("Deployment has no target revision for session creation")

        # Prepare mount spec
        mount_spec = target_revision.mounts.to_mount_spec()

        # Prepare environment variables
        environ = target_revision.execution.environ or {}
        if "BACKEND_MODEL_NAME" not in environ:
            # Add model name to environment if not already present
            environ["BACKEND_MODEL_NAME"] = deployment_info.metadata.name

        # Create kernel specs for cluster
        DEFAULT_ROLE = "main"
        kernel_specs = []
        for idx in range(target_revision.resource_spec.cluster_size):
            kernel_spec = KernelEnqueueingConfig(
                image_ref=context.image.ref,
                cluster_role=DEFAULT_ROLE if idx == 0 else "worker",
                cluster_idx=idx + 1,
                local_rank=idx,
                cluster_hostname=f"{DEFAULT_ROLE}{idx + 1}" if idx == 0 else f"worker{idx}",
                creation_config={
                    "mounts": mount_spec.mounts,
                    "mount_map": mount_spec.mount_map,
                    "mount_options": mount_spec.mount_options,
                    "environ": environ,
                    "resources": target_revision.resource_spec.resource_slots,
                    "resource_opts": target_revision.resource_spec.resource_opts,
                },
                uid=context.container_user.uid,
                main_gid=context.container_user.main_gid,
                supplementary_gids=context.container_user.supplementary_gids,
                bootstrap_script=target_revision.execution.bootstrap_script or "",
                startup_command=target_revision.execution.startup_command,
            )
            kernel_specs.append(kernel_spec)

        return cls(
            session_creation_id=session_creation_id,
            session_name=f"{deployment_info.metadata.name}-{str(route_id)}",
            access_key=context.session_owner.access_key,
            user_scope=UserScope(
                domain_name=deployment_info.metadata.domain,
                group_id=context.group_id,
                user_uuid=context.session_owner.uuid,
                user_role=context.session_owner.role,
            ),
            session_type=SessionTypes.INFERENCE,
            cluster_mode=target_revision.resource_spec.cluster_mode,
            cluster_size=target_revision.resource_spec.cluster_size,
            priority=SESSION_PRIORITY_DEFAULT,
            resource_policy=context.resource_policy,
            kernel_specs=kernel_specs,
            creation_spec={
                "mounts": mount_spec.mounts,
                "mount_map": mount_spec.mount_map,
                "mount_options": mount_spec.mount_options,
                "model_definition_path": target_revision.mounts.model_definition_path,
                "runtime_variant": target_revision.execution.runtime_variant,
                "environ": environ,
                "scaling_group": deployment_info.metadata.resource_group,
                "resources": target_revision.resource_spec.resource_slots,
                "resource_opts": target_revision.resource_spec.resource_opts,
                "preopen_ports": None,
                "agent_list": None,
            },
            scaling_group=deployment_info.metadata.resource_group,
            session_tag=deployment_info.metadata.tag,
            callback_url=target_revision.execution.callback_url,
            route_id=route_id,
            sudo_session_enabled=context.session_owner.sudo_session_enabled,
        )


@dataclass
class KernelEnqueueData:
    """Data for each kernel in the session."""

    id: KernelId
    session_id: SessionId
    session_creation_id: str
    session_name: str
    session_type: SessionTypes
    cluster_mode: str  # Store as string for DB
    cluster_size: int
    cluster_role: str
    cluster_idx: int
    local_rank: int
    cluster_hostname: str
    scaling_group: str
    domain_name: str
    group_id: UUID
    user_uuid: UUID
    access_key: AccessKey
    image: str  # Canonical image name
    architecture: str
    registry: str
    tag: Optional[str]
    starts_at: Optional[datetime]
    status: str  # KernelStatus.PENDING
    status_history: dict[str, str]
    occupied_slots: ResourceSlot
    requested_slots: ResourceSlot
    occupied_shares: dict
    resource_opts: dict
    environ: list[str]  # List of "KEY=VALUE" strings
    bootstrap_script: Optional[str]
    startup_command: Optional[str]
    internal_data: dict[str, Any]
    callback_url: Optional[yarl.URL]
    mounts: list[str]  # Legacy field for compatibility
    vfolder_mounts: list[VFolderMount]
    preopen_ports: list[int]
    use_host_network: bool

    # Port fields (initially 0)
    repl_in_port: int = 0
    repl_out_port: int = 0
    stdin_port: int = 0
    stdout_port: int = 0

    # Container user info
    uid: Optional[int] = field(default=None)
    main_gid: Optional[int] = field(default=None)
    gids: list[int] = field(default_factory=list)


@dataclass
class SessionEnqueueData:
    """Prepared data ready to be enqueued in database."""

    # Associated data
    kernels: list[KernelEnqueueData]
    dependencies: list[SessionId]

    id: SessionId
    creation_id: str
    name: str
    access_key: AccessKey
    user_uuid: UUID
    group_id: UUID
    domain_name: str
    scaling_group_name: str
    session_type: SessionTypes
    cluster_mode: str  # Store as string for DB
    cluster_size: int
    priority: int
    status: str  # SessionStatus.PENDING
    status_history: dict[str, str]
    requested_slots: ResourceSlot
    occupying_slots: ResourceSlot
    vfolder_mounts: list[VFolderMount]
    environ: dict[str, str]
    tag: Optional[str]
    starts_at: Optional[datetime]
    batch_timeout: Optional[int]  # seconds
    callback_url: Optional[yarl.URL]
    images: list[str]
    designated_agent_list: Optional[list[str]]
    network_type: Optional[NetworkType] = None
    network_id: Optional[str] = None
    bootstrap_script: Optional[str] = None
    use_host_network: bool = False
    timeout: Optional[int] = None


@dataclass
class SessionDependencyData:
    """Data for session dependency relationships."""

    session_id: SessionId
    depends_on: SessionId


@dataclass
class ScalingGroupNetworkInfo:
    """Network configuration from scaling group."""

    use_host_network: bool
    wsproxy_addr: Optional[str] = None


@dataclass
class ImageInfo:
    """Resolved image information."""

    canonical: str
    architecture: str
    registry: str
    labels: dict[str, Any]
    # Resource spec maps slot names to {"min": value, "max": value}
    # Values can be strings (for BinarySize), numbers, or None
    resource_spec: dict[str, dict[str, Optional[str | int | Decimal]]]


@dataclass
class AllowedScalingGroup:
    """Allowed scaling group for a user."""

    name: str
    is_private: bool
    scheduler_opts: ScalingGroupOpts


@dataclass
class ContainerUserInfo:
    """User container UID/GID information."""

    uid: Optional[int] = None
    main_gid: Optional[int] = None
    supplementary_gids: list[int] = field(default_factory=list)


@dataclass
class SessionCreationContext:
    """All data needed for session creation, fetched in a single batch."""

    # Scaling group network info
    scaling_group_network: ScalingGroupNetworkInfo

    # Allowed scaling groups for validation
    allowed_scaling_groups: list[AllowedScalingGroup]

    # Image information for all kernels
    image_infos: dict[str, ImageInfo]  # keyed by image reference

    # Vfolder mounts
    vfolder_mounts: list[VFolderMount]

    # Dotfile data
    dotfile_data: dict[str, Any]

    # User (UID/GID) inside container
    container_user_info: ContainerUserInfo



from typing import Any, Mapping, Optional, Sequence
from uuid import UUID

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class VFolderNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/vfolder-not-found"
    error_title = "Virtual folder not found."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class VFolderInvitationNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/vfolder-invitation-not-found"
    error_title = "Virtual folder invitation not found."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER_INVITATION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class VFolderCreationFailure(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-creation-failed"
    error_title = "Virtual folder creation failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class VFolderAlreadyExists(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/vfolder-already-exists"
    error_title = "Virtual folder already exists."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class VFolderGrantAlreadyExists(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/vfolder-already-exists"
    error_title = "Virtual folder already exists."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.GRANT,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class VFolderInvalidParameter(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-parameter"
    error_title = "Invalid parameter."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InsufficientPrivilege(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/insufficient-privilege"
    error_title = "Insufficient privilege."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class Forbidden(VFolderInvalidParameter, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/forbidden"
    error_title = "Forbidden operation."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class ProjectNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/project-not-found"
    error_title = "Project not found."

    _project_id: Optional[str | UUID]

    def __init__(self, project_id: Optional[str | UUID]) -> None:
        self._project_id = project_id

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.GROUP,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class InternalServerError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Internal server error."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ModelServiceDependencyNotCleared(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/model-service-dependency-not-cleared"
    error_title = "Cannot delete model VFolders bound to alive model services."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class TooManyVFoldersFound(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/too-many-vfolders"
    error_title = "Too many virtual folders found."

    targets: Sequence[Mapping[str, Any]]

    def __init__(self, targets: Sequence[Mapping[str, Any]]) -> None:
        self.targets = targets

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )


class VFolderFilterStatusNotAvailable(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-filter-status-not-available"
    error_title = "Virtual folder filter status not available."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class VFolderFilterStatusFailed(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/vfolder-filter-status-failed"
    error_title = "Virtual folder filter status failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )



from __future__ import annotations

import uuid
from typing import Any, Optional
from uuid import UUID

import graphene
import graphql
from graphene.types import Scalar
from graphene.types.scalars import MAX_INT, MIN_INT
from graphql import GraphQLError
from graphql.language.ast import FloatValueNode, IntValueNode, ObjectValueNode, ValueNode
from graphql.language.printer import print_ast

from ...errors.common import ObjectNotFound
from ..gql_relay import AsyncNode

SAFE_MIN_INT = -9007199254740991
SAFE_MAX_INT = 9007199254740991


class ResourceLimit(graphene.ObjectType):
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


class KVPair(graphene.ObjectType):
    key = graphene.String()
    value = graphene.String()


class ResourceLimitInput(graphene.InputObjectType):
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


class KVPairInput(graphene.InputObjectType):
    key = graphene.String()
    value = graphene.String()


class BigInt(Scalar):
    """
    BigInt is an extension of the regular graphene.Int scalar type
    to support integers outside the range of a signed 32-bit integer.
    """

    @staticmethod
    def coerce_bigint(value):
        num = int(value)
        if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
            raise ValueError("Cannot serialize integer out of the safe range.")
        if not (MIN_INT <= num <= MAX_INT):
            # treat as float
            return float(int(num))
        return num

    serialize = coerce_bigint
    parse_value = coerce_bigint

    @staticmethod
    def parse_literal(node):
        if isinstance(node, IntValueNode):
            num = int(node.value)
            if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
                raise ValueError("Cannot parse integer out of the safe range.")
            if not (MIN_INT <= num <= MAX_INT):
                # treat as float
                return float(int(num))
            return num


class Bytes(Scalar):
    class Meta:
        description = "Added in 24.09.1."

    @staticmethod
    def serialize(val: bytes) -> str:
        return val.hex()

    @staticmethod
    def parse_literal(node: Any, _variables=None) -> Optional[bytes]:
        if isinstance(node, graphql.language.ast.StringValueNode):
            assert isinstance(node, str)
            return bytes.fromhex(node)
        return None

    @staticmethod
    def parse_value(value: str) -> bytes:
        return bytes.fromhex(value)


class ImageRefType(graphene.InputObjectType):
    name = graphene.String(required=True)
    registry = graphene.String()
    architecture = graphene.String()


class UUIDFloatMap(Scalar):
    """
    Added in 25.4.0.
    Verifies that the key is a UUID (represented as a string) and the value is a float.
    """

    @staticmethod
    def serialize(value: Any) -> dict[str, float]:
        if not isinstance(value, dict):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-dict value: {repr(value)}")

        validated: dict[str, float] = {}
        for k, v in value.items():
            try:
                key_str = str(uuid.UUID(k))
            except ValueError:
                raise GraphQLError(f"UUIDFloatMap cannot represent key {k} as a valid UUID")

            if not isinstance(v, float):
                raise GraphQLError(f"UUIDFloatMap cannot represent value {v} as a float")
            validated[key_str] = v
        return validated

    @classmethod
    def parse_literal(cls, node: ValueNode, _variables=None) -> dict[str, float]:
        if not isinstance(node, ObjectValueNode):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-object value: {print_ast(node)}")
        validated: dict[str, Any] = {}
        for field in node.fields:
            key = field.name.value
            if isinstance(field.value, (FloatValueNode, IntValueNode)):
                try:
                    validated[key] = float(field.value.value)
                except Exception:
                    raise GraphQLError(
                        f"UUIDFloatMap cannot represent value for key {key} as a float"
                    )
            else:
                raise GraphQLError(
                    f"UUIDFloatMap cannot represent non-numeric value for key {key}: {print_ast(field.value)}"
                )
        return validated

    @staticmethod
    def parse_value(value: Any) -> dict[str, float]:
        if not isinstance(value, dict):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-dict value: {repr(value)}")
        validated: dict[str, float] = {}
        for k, v in value.items():
            try:
                key_str = str(uuid.UUID(k))
            except ValueError:
                raise GraphQLError(f"UUIDFloatMap cannot represent key {k} as a valid UUID")
            if not isinstance(v, float):
                raise GraphQLError(f"UUIDFloatMap cannot represent value {v} as a float")
            validated[key_str] = v
        return validated


def extract_object_uuid(info: graphene.ResolveInfo, global_id: str, object_name: str) -> UUID:
    """
    Converts a GraphQL global ID to its corresponding UUID.
    If the global ID is not valid, raises an error using the provided object name.
    """

    _, raw_id = AsyncNode.resolve_global_id(info, global_id)
    if not raw_id:
        raw_id = global_id

    try:
        return UUID(raw_id)
    except ValueError:
        raise ObjectNotFound(object_name)



import uuid
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Any, Optional, override

import yarl
from pydantic import AnyUrl

from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    RuntimeVariant,
    VFolderMount,
)
from ai.backend.manager.data.model_serving.types import (
    EndpointLifecycle,
    ModelServicePrepareCtx,
    ServiceConfig,
)
from ai.backend.manager.types import Creator


@dataclass
class EndpointCreator(Creator):
    name: str
    model_definition_path: Optional[str]
    created_user: uuid.UUID
    session_owner: uuid.UUID
    image: uuid.UUID  # Image row ID
    model: uuid.UUID  # vfolder row ID
    domain: str
    project: uuid.UUID
    resource_group: str  # Resource group row ID which is the name
    resource_slots: Mapping[str, Any]
    replicas: int = 0
    lifecycle_stage: EndpointLifecycle = EndpointLifecycle.CREATED
    tag: Optional[str] = None
    startup_command: Optional[str] = None
    bootstrap_script: Optional[str] = None
    callback_url: Optional[yarl.URL] = None
    environ: Optional[dict[str, str]] = None
    open_to_public: bool = False
    runtime_variant: RuntimeVariant = RuntimeVariant.CUSTOM
    model_mount_destination: str = "/models"
    url: Optional[str] = None
    resource_opts: Optional[dict[str, Any]] = None
    cluster_mode: ClusterMode = ClusterMode.SINGLE_NODE
    cluster_size: int = 1
    extra_mounts: list[VFolderMount] = field(default_factory=list)
    retries: int = 0

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "model_definition_path": self.model_definition_path,
            "created_user": self.created_user,
            "session_owner": self.session_owner,
            "image": self.image,
            "model": self.model,
            "domain": self.domain,
            "project": self.project,
            "resource_group": self.resource_group,
            "resource_slots": self.resource_slots,
            "replicas": self.replicas,
            "lifecycle_stage": self.lifecycle_stage,
            "tag": self.tag,
            "startup_command": self.startup_command,
            "bootstrap_script": self.bootstrap_script,
            "callback_url": self.callback_url,
            "environ": self.environ,
            "open_to_public": self.open_to_public,
            "runtime_variant": self.runtime_variant,
            "model_mount_destination": self.model_mount_destination,
            "url": self.url,
            "resource_opts": self.resource_opts,
            "cluster_mode": self.cluster_mode,
            "cluster_size": self.cluster_size,
            "extra_mounts": self.extra_mounts,
        }


@dataclass
class ModelServiceCreator(Creator):
    service_name: str
    replicas: int
    image: str
    runtime_variant: RuntimeVariant
    architecture: str
    group_name: str
    domain_name: str
    cluster_size: int
    cluster_mode: ClusterMode
    open_to_public: bool
    config: ServiceConfig
    sudo_session_enabled: bool
    model_service_prepare_ctx: ModelServicePrepareCtx
    tag: Optional[str] = None
    startup_command: Optional[str] = None
    bootstrap_script: Optional[str] = None
    callback_url: Optional[AnyUrl] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {}


@dataclass
class EndpointAutoScalingRuleCreator(Creator):
    metric_source: AutoScalingMetricSource
    metric_name: str
    threshold: str
    comparator: AutoScalingMetricComparator
    step_size: int
    cooldown_seconds: int
    min_replicas: Optional[int] = None
    max_replicas: Optional[int] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "metric_source": self.metric_source,
            "metric_name": self.metric_name,
            "threshold": self.threshold,
            "comparator": self.comparator,
            "step_size": self.step_size,
            "cooldown_seconds": self.cooldown_seconds,
            "min_replicas": self.min_replicas,
            "max_replicas": self.max_replicas,
        }



"""Add `remote_status` to `artifact_revision`

Revision ID: fa3dd7f77c19
Revises: 82f4d3dea750
Create Date: 2025-09-30 10:49:45.228649

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fa3dd7f77c19"
down_revision = "82f4d3dea750"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "artifact_revisions",
        sa.Column("remote_status", sa.String(), nullable=True, server_default=sa.null()),
    )
    op.create_index(
        op.f("ix_artifact_revisions_remote_status"),
        "artifact_revisions",
        ["remote_status"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_artifact_revisions_remote_status"), table_name="artifact_revisions")
    op.drop_column("artifact_revisions", "remote_status")
    # ### end Alembic commands ###



import dataclasses
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional
from uuid import UUID

from ai.backend.manager.data.session.types import SessionStatus


@dataclass
class LegacySessionInfo:
    """
    Deprecated.
    Only use this type in `GetSessionInfo` Action.
    """

    age_ms: int = field(compare=False)
    domain_name: str
    group_id: UUID
    user_id: UUID
    lang: str
    image: str
    architecture: str
    registry: Optional[str]
    tag: Optional[str]
    container_id: UUID
    occupied_slots: str  # legacy
    occupying_slots: str
    requested_slots: str
    occupied_shares: str  # legacy
    environ: str
    resource_opts: str
    status: SessionStatus
    status_info: Optional[str]
    status_data: Optional[dict[str, Any]]
    creation_time: datetime
    termination_time: Optional[datetime]
    num_queries_executed: int
    last_stat: Optional[dict[str, Any]]
    idle_checks: Any

    def asdict(self) -> dict[str, Any]:
        """
        Return a dict whose keys exactly match the old `resp` payload.
        """
        return {
            "age": self.age_ms,
            "domainName": self.domain_name,
            "groupId": str(self.group_id),
            "userId": str(self.user_id),
            "lang": self.lang,
            "image": self.image,
            "architecture": self.architecture,
            "registry": self.registry,
            "tag": self.tag,
            "containerId": str(self.container_id),
            "occupiedSlots": self.occupied_slots,
            "occupyingSlots": self.occupying_slots,
            "requestedSlots": self.requested_slots,
            "occupiedShares": self.occupied_shares,
            "environ": self.environ,
            "resourceOpts": self.resource_opts,
            "status": self.status.name,  # Enum → its name (e.g. "RUNNING")
            "statusInfo": self.status_info or "None",  # legacy
            "statusData": self.status_data,
            "creationTime": self.creation_time.isoformat(),
            "terminationTime": (
                self.termination_time.isoformat() if self.termination_time else None
            ),
            "numQueriesExecuted": self.num_queries_executed,
            "lastStat": self.last_stat,
            "idleChecks": self.idle_checks,
        }


@dataclass
class CommitStatusInfo:
    status: str
    kernel: str

    def asdict(self) -> dict[str, Any]:
        return dataclasses.asdict(self)



from dataclasses import dataclass
from typing import Any, override

from ai.backend.manager.types import Creator


@dataclass
class ReservoirRegistryCreator(Creator):
    endpoint: str
    access_key: str
    secret_key: str
    api_version: str

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "endpoint": self.endpoint,
            "access_key": self.access_key,
            "secret_key": self.secret_key,
            "api_version": self.api_version,
        }



import base64
from datetime import datetime, timedelta
from enum import StrEnum
from typing import AsyncGenerator, Optional, cast
from uuid import uuid4

import strawberry
from strawberry import ID, Info, relay
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo
from strawberry.relay.types import NodeIterableType

from ai.backend.manager.api.gql.base import JSONString, OrderDirection, StringFilter
from ai.backend.manager.api.gql.federated_types import (
    AccessToken,
    AutoScalingRule,
    ResourceGroup,
    User,
)
from ai.backend.manager.api.gql.model_deployment.routing import (
    RoutingNode,
)

from .model_revision import (
    CreateModelRevisionInput,
    ModelRevision,
    ModelRevisionConnection,
    ModelRevisionEdge,
    mock_model_revision_1,
    mock_model_revision_2,
    mock_model_revision_3,
)


@strawberry.enum
class ClusterMode(StrEnum):
    SINGLE_NODE = "SINGLE_NODE"
    MULTI_NODE = "MULTI_NODE"


@strawberry.enum
class DeploymentStatus(StrEnum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"


@strawberry.enum
class ReplicaStatus(StrEnum):
    HEALTHY = "HEALTHY"
    UNHEALTHY = "UNHEALTHY"


@strawberry.enum
class DeploymentStrategyType(StrEnum):
    ROLLING = "ROLLING"
    BLUE_GREEN = "BLUE_GREEN"
    CANARY = "CANARY"


@strawberry.enum
class DeploymentOrderField(StrEnum):
    CREATED_AT = "CREATED_AT"
    UPDATED_AT = "UPDATED_AT"
    NAME = "NAME"


@strawberry.type
class DeploymentStrategy:
    type: DeploymentStrategyType


@strawberry.type
class ModelReplica(Node):
    id: NodeID
    name: str
    status: ReplicaStatus
    revision: ModelRevision
    routings: list[RoutingNode]


@strawberry.type
class ModelReplicaConnection(Connection[ModelReplica]):
    @strawberry.field
    def count(self) -> int:
        return 0

    @classmethod
    def resolve_connection(
        cls,
        nodes: NodeIterableType[ModelReplica],
        *,
        info: Info,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        max_results: Optional[int] = None,
        **kwargs,
    ) -> "ModelReplicaConnection":
        return cls(
            edges=[],
            page_info=relay.PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        )


@strawberry.type
class ClusterConfig:
    mode: ClusterMode
    size: int


@strawberry.type
class ReplicaState:
    desired_replica_count: int
    replicas: ModelReplicaConnection


@strawberry.type
class ScalingRule:
    auto_scaling_rules: list[AutoScalingRule]


@strawberry.type
class ModelDeploymentMetadata:
    name: str
    status: DeploymentStatus
    tags: list[str]
    created_at: datetime
    updated_at: datetime


@strawberry.type
class ModelDeploymentNetworkAccess:
    endpoint_url: Optional[str] = None
    preferred_domain_name: Optional[str] = None
    open_to_public: bool = False
    access_tokens: list[AccessToken]


@strawberry.type
class ResourceConfig:
    resource_group: ResourceGroup
    resource_slots: JSONString
    resource_opts: Optional[JSONString] = None


# Main ModelDeployment Type
@strawberry.type
class ModelDeployment(Node):
    id: NodeID
    metadata: ModelDeploymentMetadata
    network_access: ModelDeploymentNetworkAccess

    revision: Optional[ModelRevision] = None
    revision_history: ModelRevisionConnection

    scaling_rule: ScalingRule
    replica_state: ReplicaState

    deployment_strategy: DeploymentStrategy

    cluster_config: ClusterConfig
    resource_config: ResourceConfig
    created_user: User


# Filter Types
@strawberry.input
class DeploymentFilter:
    status: Optional[DeploymentStatus] = None
    open_to_public: Optional[bool] = None
    tags: Optional[list[StringFilter]] = None

    AND: Optional["DeploymentFilter"] = None
    OR: Optional["DeploymentFilter"] = None
    NOT: Optional["DeploymentFilter"] = None
    DISTINCT: Optional[bool] = None


@strawberry.input
class DeploymentOrderBy:
    field: DeploymentOrderField
    direction: OrderDirection = OrderDirection.DESC


# Payload Types
@strawberry.type
class CreateModelDeploymentPayload:
    deployment: ModelDeployment


@strawberry.type
class UpdateModelDeploymentPayload:
    deployment: Optional[ModelDeployment] = None


@strawberry.type
class DeleteModelDeploymentPayload:
    deployment: Optional[ModelDeployment] = None


@strawberry.type
class DeploymentStatusChangedPayload:
    deployment: ModelDeployment


@strawberry.type
class ReplicaStatusChangedPayload:
    replica: ModelReplica


# Input Types
@strawberry.input
class ClusterConfigInput:
    mode: ClusterMode
    size: int


@strawberry.input
class ResourceGroupInput:
    name: str


@strawberry.input
class ResourceConfigInput:
    resource_group: ResourceGroupInput
    resource_slots: JSONString
    resource_opts: Optional[JSONString] = None


@strawberry.input
class ModelDeploymentMetadataInput:
    name: str
    tags: Optional[list[str]] = None


@strawberry.input
class ModelDeploymentNetworkAccessInput:
    preferred_domain_name: Optional[str] = None
    open_to_public: bool = False


@strawberry.input
class DeploymentStrategyInput:
    type: DeploymentStrategyType


@strawberry.input
class CreateModelDeploymentInput:
    metadata: ModelDeploymentMetadataInput
    network_access: ModelDeploymentNetworkAccessInput
    cluster_config: ClusterConfigInput
    resource_config: ResourceConfigInput
    deployment_strategy: DeploymentStrategyInput
    initial_revision: CreateModelRevisionInput


@strawberry.input
class UpdateModelDeploymentInput:
    id: ID
    open_to_public: Optional[bool] = None
    tags: Optional[list[str]] = None
    deployment_strategy: Optional[DeploymentStrategyInput] = None
    active_revision_id: Optional[ID] = None


@strawberry.input
class DeleteModelDeploymentInput:
    id: ID


def _generate_mock_global_id() -> str:
    return base64.b64encode(f"default:{uuid4()}".encode("utf-8")).decode()


# Mock Model Replicas
mock_model_replica_1 = ModelReplica(
    id=_generate_mock_global_id(),
    name="llama-3-8b-instruct-replica-01",
    status=ReplicaStatus.HEALTHY,
    revision=mock_model_revision_1,
    routings=[
        RoutingNode(
            id=ID(_generate_mock_global_id()),
            routing_id=uuid4(),
            endpoint="https://api.backend.ai/models/dep-001/routing/01",
            session=uuid4(),
            status="ACTIVE",
            traffic_ratio=0.33,
            created_at=datetime.now() - timedelta(days=5),
            live_stat=cast(
                JSONString, '{"requests": 1523, "latency_ms": 187, "tokens_per_second": 42.5}'
            ),
        )
    ],
)

mock_model_replica_2 = ModelReplica(
    id=_generate_mock_global_id(),
    name="llama-3-8b-instruct-replica-02",
    status=ReplicaStatus.HEALTHY,
    revision=mock_model_revision_1,
    routings=[
        RoutingNode(
            id=ID(_generate_mock_global_id()),
            routing_id=uuid4(),
            endpoint="https://api.backend.ai/models/dep-001/routing/02",
            session=uuid4(),
            status="ACTIVE",
            traffic_ratio=0.33,
            created_at=datetime.now() - timedelta(days=5),
            live_stat=cast(
                JSONString, '{"requests": 1456, "latency_ms": 195, "tokens_per_second": 41.2}'
            ),
        )
    ],
)

mock_model_replica_3 = ModelReplica(
    id=_generate_mock_global_id(),
    name="llama-3-8b-instruct-replica-03",
    status=ReplicaStatus.UNHEALTHY,
    revision=mock_model_revision_1,
    routings=[
        RoutingNode(
            id=ID(_generate_mock_global_id()),
            routing_id=uuid4(),
            endpoint="https://api.backend.ai/models/dep-001/routing/03",
            session=uuid4(),
            status="INACTIVE",
            traffic_ratio=0.0,
            created_at=datetime.now() - timedelta(days=2),
            live_stat=cast(JSONString, '{"requests": 0, "latency_ms": 0, "tokens_per_second": 0}'),
        )
    ],
)

ModelReplicaEdge = Edge[ModelReplica]

# TODO: After implementing the actual logic, remove these mock objects
# Mock Model Deployments
mock_model_deployment_1 = ModelDeployment(
    id=_generate_mock_global_id(),
    metadata=ModelDeploymentMetadata(
        name="Llama 3.8B Instruct",
        status=DeploymentStatus.ACTIVE,
        tags=["production", "llm", "chat", "instruct"],
        created_at=datetime.now() - timedelta(days=30),
        updated_at=datetime.now() - timedelta(hours=2),
    ),
    network_access=ModelDeploymentNetworkAccess(
        endpoint_url="https://api.backend.ai/models/dep-001",
        preferred_domain_name="llama-3-8b.models.backend.ai",
        open_to_public=True,
        access_tokens=[],
    ),
    cluster_config=ClusterConfig(mode=ClusterMode.SINGLE_NODE, size=1),
    resource_config=ResourceConfig(
        resource_group=ResourceGroup(id=ID(_generate_mock_global_id())),
        resource_slots=cast(
            JSONString,
            '{"cpu": 8, "mem": "32G", "cuda.shares": 1, "cuda.device": 1}',
        ),
        resource_opts=cast(
            JSONString,
            '{"shmem": "2G", "reserved_time": "24h", "scaling_group": "us-east-1"}',
        ),
    ),
    revision=mock_model_revision_1,
    revision_history=ModelRevisionConnection(
        edges=[
            ModelRevisionEdge(node=mock_model_revision_1, cursor="rev-cursor-1"),
            ModelRevisionEdge(node=mock_model_revision_2, cursor="rev-cursor-2"),
        ],
        page_info=PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor="rev-cursor-1",
            end_cursor="rev-cursor-2",
        ),
    ),
    scaling_rule=ScalingRule(
        auto_scaling_rules=[
            AutoScalingRule(id=ID(_generate_mock_global_id())),
            AutoScalingRule(id=ID(_generate_mock_global_id())),
        ]
    ),
    replica_state=ReplicaState(
        desired_replica_count=3,
        replicas=ModelReplicaConnection(
            edges=[
                ModelReplicaEdge(node=mock_model_replica_1, cursor="replica-cursor-1"),
                ModelReplicaEdge(node=mock_model_replica_2, cursor="replica-cursor-2"),
                ModelReplicaEdge(node=mock_model_replica_3, cursor="replica-cursor-3"),
            ],
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor="replica-cursor-1",
                end_cursor="replica-cursor-3",
            ),
        ),
    ),
    deployment_strategy=DeploymentStrategy(type=DeploymentStrategyType.ROLLING),
    created_user=User(id=ID(_generate_mock_global_id())),
)

mock_model_deployment_2 = ModelDeployment(
    id=_generate_mock_global_id(),
    metadata=ModelDeploymentMetadata(
        name="Mistral 7B v0.3",
        status=DeploymentStatus.ACTIVE,
        tags=["staging", "llm", "experimental"],
        created_at=datetime.now() - timedelta(days=20),
        updated_at=datetime.now() - timedelta(days=1),
    ),
    network_access=ModelDeploymentNetworkAccess(
        endpoint_url="https://api.backend.ai/models/dep-002",
        preferred_domain_name="mistral-7b.models.backend.ai",
        open_to_public=False,
        access_tokens=[],
    ),
    cluster_config=ClusterConfig(mode=ClusterMode.SINGLE_NODE, size=1),
    resource_config=ResourceConfig(
        resource_group=ResourceGroup(id=ID(_generate_mock_global_id())),
        resource_slots=cast(
            JSONString,
            '{"cpu": 8, "mem": "32G", "cuda.shares": 1, "cuda.device": 1}',
        ),
        resource_opts=cast(
            JSONString,
            '{"shmem": "2G", "reserved_time": "24h", "scaling_group": "us-east-1"}',
        ),
    ),
    revision=mock_model_revision_3,
    revision_history=ModelRevisionConnection(
        edges=[
            ModelRevisionEdge(node=mock_model_revision_3, cursor="rev-cursor-3"),
        ],
        page_info=PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor="rev-cursor-3",
            end_cursor="rev-cursor-3",
        ),
    ),
    scaling_rule=ScalingRule(auto_scaling_rules=[]),
    replica_state=ReplicaState(
        desired_replica_count=1,
        replicas=ModelReplicaConnection(
            edges=[
                ModelReplicaEdge(node=mock_model_replica_1, cursor="replica-cursor-1"),
                ModelReplicaEdge(node=mock_model_replica_2, cursor="replica-cursor-2"),
            ],
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        ),
    ),
    deployment_strategy=DeploymentStrategy(type=DeploymentStrategyType.BLUE_GREEN),
    created_user=User(id=ID(_generate_mock_global_id())),
)

mock_model_deployment_3 = ModelDeployment(
    id=_generate_mock_global_id(),
    metadata=ModelDeploymentMetadata(
        name="Gemma 2.9B",
        status=DeploymentStatus.INACTIVE,
        tags=["development", "llm", "testing"],
        created_at=datetime.now() - timedelta(days=15),
        updated_at=datetime.now() - timedelta(days=7),
    ),
    network_access=ModelDeploymentNetworkAccess(
        endpoint_url=None,
        preferred_domain_name=None,
        open_to_public=False,
        access_tokens=[],
    ),
    cluster_config=ClusterConfig(mode=ClusterMode.SINGLE_NODE, size=1),
    resource_config=ResourceConfig(
        resource_group=ResourceGroup(id=ID(_generate_mock_global_id())),
        resource_slots=cast(
            JSONString,
            '{"cpu": 8, "mem": "32G", "cuda.shares": 1, "cuda.device": 1}',
        ),
        resource_opts=cast(
            JSONString,
            '{"shmem": "2G", "reserved_time": "24h", "scaling_group": "us-east-1"}',
        ),
    ),
    revision=None,
    revision_history=ModelRevisionConnection(
        edges=[],
        page_info=PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=None,
            end_cursor=None,
        ),
    ),
    scaling_rule=ScalingRule(auto_scaling_rules=[]),
    replica_state=ReplicaState(
        desired_replica_count=0,
        replicas=ModelReplicaConnection(
            edges=[],
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        ),
    ),
    deployment_strategy=DeploymentStrategy(type=DeploymentStrategyType.CANARY),
    created_user=User(id=ID(_generate_mock_global_id())),
)


ModelDeploymentEdge = Edge[ModelDeployment]


# Connection types for Relay support
@strawberry.type
class ModelDeploymentConnection(Connection[ModelDeployment]):
    @strawberry.field
    def count(self) -> int:
        return 0

    @classmethod
    def resolve_connection(
        cls,
        nodes: NodeIterableType[ModelDeployment],
        *,
        info: Info,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        max_results: Optional[int] = None,
        **kwargs,
    ) -> "ModelDeploymentConnection":
        mock_deployments = [
            mock_model_deployment_1,
            mock_model_deployment_2,
            mock_model_deployment_3,
        ]
        return cls(
            edges=[
                Edge(node=deployment, cursor=str(i))
                for i, deployment in enumerate(mock_deployments)
            ],
            page_info=relay.PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        )


# Resolvers
@strawberry.relay.connection(ModelDeploymentConnection)
async def deployments(
    filter: Optional[DeploymentFilter] = None,
    order_by: Optional[DeploymentOrderBy] = None,
    first: Optional[int] = None,
    after: Optional[str] = None,
) -> list[ModelDeployment]:
    """List deployments with optional filtering and pagination."""
    # Return a list of mock deployments with more details
    deployments = [
        mock_model_deployment_1,
        mock_model_deployment_2,
        mock_model_deployment_3,
    ]

    return deployments


@strawberry.field
async def deployment(id: ID) -> Optional[ModelDeployment]:
    """Get a specific deployment by ID."""
    return None


@strawberry.field
async def replica(id: ID) -> Optional[ModelReplica]:
    """Get a specific replica by ID."""

    return ModelReplica(
        id=id,
        name="llama-3-8b-instruct-replica-01",
        status=ReplicaStatus.HEALTHY,
        revision=mock_model_revision_1,
        routings=[],
    )


@strawberry.mutation
async def create_model_deployment(
    input: CreateModelDeploymentInput,
) -> CreateModelDeploymentPayload:
    """Create a new model deployment."""
    # Create a dummy deployment for placeholder
    return CreateModelDeploymentPayload(deployment=mock_model_deployment_1)


@strawberry.mutation
async def update_model_deployment(
    input: UpdateModelDeploymentInput,
) -> UpdateModelDeploymentPayload:
    """Update an existing model deployment."""
    # Create a dummy deployment for placeholder
    return UpdateModelDeploymentPayload(deployment=mock_model_deployment_1)


@strawberry.mutation
async def delete_model_deployment(
    input: DeleteModelDeploymentInput,
) -> DeleteModelDeploymentPayload:
    """Delete a model deployment."""
    return DeleteModelDeploymentPayload(deployment=None)


@strawberry.subscription
async def deployment_status_changed(
    deployment_id: ID,
) -> AsyncGenerator[DeploymentStatusChangedPayload, None]:
    """Subscribe to deployment status changes."""
    deployment = [mock_model_deployment_1, mock_model_deployment_2, mock_model_deployment_3]

    for dep in deployment:
        yield DeploymentStatusChangedPayload(deployment=dep)


@strawberry.subscription
async def replica_status_changed(
    revision_id: ID,
) -> AsyncGenerator[ReplicaStatusChangedPayload, None]:
    """Subscribe to replica status changes."""
    replicas = [mock_model_replica_1, mock_model_replica_2, mock_model_replica_3]

    for replica in replicas:
        yield ReplicaStatusChangedPayload(replica=replica)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryCreatorMeta
from ai.backend.manager.data.huggingface_registry.creator import HuggingFaceRegistryCreator
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class CreateHuggingFaceRegistryAction(ArtifactRegistryAction):
    creator: HuggingFaceRegistryCreator
    meta: ArtifactRegistryCreatorMeta

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create_huggingface_registry"


@dataclass
class CreateHuggingFaceRegistryActionResult(BaseActionResult):
    result: HuggingFaceRegistryData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.result.id)



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.session.base import SessionAction


# TODO: Change this to BatchAction since it can destroy multiple sessions with recursive option
@dataclass
class DestroySessionAction(SessionAction):
    user_role: UserRole
    session_name: str
    forced: bool
    recursive: bool
    owner_access_key: AccessKey

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        # TODO: Handle this
        # if self.recursive:
        #     return "destory_multi"
        return "destory"


@dataclass
class DestroySessionActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any

    # TODO: Change this to `entity_ids`
    @override
    def entity_id(self) -> Optional[str]:
        return None



"""add scope entity mapping table

Revision ID: ed84197be4fe
Revises: a5ef73b01e97
Create Date: 2025-07-11 22:16:00.414113

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "ed84197be4fe"
down_revision = "a5ef73b01e97"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "association_scopes_entities",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("scope_type", sa.String(length=32), nullable=False),
        sa.Column("scope_id", sa.String(length=64), nullable=False),
        sa.Column("entity_type", sa.String(length=32), nullable=False),
        sa.Column("entity_id", sa.String(length=64), nullable=False),
        sa.Column(
            "registered_at",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_association_scopes_entities")),
        sa.UniqueConstraint("scope_id", "entity_id", name="uq_scope_id_entity_id"),
    )


def downgrade() -> None:
    op.drop_table("association_scopes_entities")
